#' Adapt dataset for Pkanalix
#'
#' Adapt the formatting of your dataset for use in PKanalix. In addition to
#' creating an adapted dataset, the function creates a PKanalix project based on
#' the adapted dataset and configured with the settings specified with
#' ncaSettings, beSettings and dataSettings, and the covariates specified with
#' catCov and contCov. See http://pkanalix.lixoft.com/data-formatting-macro/ for
#' more information
#'
#' The following actions are performed \itemize{ \item A column AMT is added,
#' containing the same dosing regimen for all subjects or subject-occasions
#' (with an option to read individual dose amounts from a column of the dataset)
#' Note that the dosing times are correct in the column tagged as TIME in the
#' form, but other time columns will not be consistent. \item If the option
#' addOccperdose is TRUE (with multiple doses), two columns Dosing_OCC and
#' Time_LD are added containing occasion indexes for the different doses, and
#' the time after the last dose, respectively. \item If the option
#' is.timeSinceLastDose is TRUE, a header for the column containing occasion
#' marks for the different doses must be given, and it is treated as an occasion
#' column. \item In case of occasions, if the occasion column does not contain
#' only numbers, a new column OCC is added with occasion numbers. \item If an
#' occasion column is also specified as Administration Id, a new column ADM_ID
#' is added with integers for different administration ids. \item If several
#' observation columns are given, they are merged into a single column OBS and a
#' column OBS_ID is created to distinguish the different types of observations.
#' \item In case of urine data, the volume is set to 0 at dosing times to avoid
#' format errors, and both start and end times of the interval are set to the
#' dosing time. The function checks that time intervals are continuous (no gaps
#' are allowed). Only the column END TIME should be used as TIME in PKanalix.
#' \item In case of censored observations, a column CENS is added, containing 0
#' for non-censored observations and 1 for censored observations. The LLOQ is
#' written instead of the censoring tag in the column of observations. \item
#' Missing observations (all values that are not numbers) are replaced by '.'.
#' \item The first lines that do not contain data are concatenated into a single
#' header row. \item Rows that do not contain data are deleted. \item Formatting
#' is adapted for correct exporting to CSV format. }
#'
#' @param dataFilename (\emph{string}) a dataset filename. It can be a txt file,
#'   a csv file, an excel xls/xlsx file.
#' @param dataType (\emph{string}) Type of data. One of {"plasma", "urine"}
#'   ("plasma" by default)
#' @param idHeader (\emph{string}) Header of the column containing ID
#'   information.
#' @param obsHeader (\emph{string} or \emph{vector}) Header name or vector of
#'   header names of the columns containing OBSERVATION information.
#' @param occHeader (\emph{string} or \emph{vector}) Header name or vector of
#'   header names of the columns containing OCCASION information. (by default no
#'   occasion is considered)
#' @param admIdHeader (\emph{string}) Header of the column containing
#'   ADMINISTRATION ID information (has to be in occHeader as well).
#' @param timeHeader (\emph{string}) Header of the column containing TIME
#'   information. To define when dataType = "plasma
#' @param urineStartTimeHeader (\emph{string}) Header of the column containing
#'   start time of urine collection intervals information. To define when
#'   dataType = "urine"
#' @param urineEndTimeHeader (\emph{string}) Header of the column containing end
#'   time of urine collection intervals information. To define when dataType =
#'   "urine"
#' @param urineVolumeHeader (\emph{string}) Header of the column containing
#'   urine collection volume information. To define when dataType = "urine"
#' @param is.multiDose (\emph{bool}) If TRUE, multiple doses / If FALSE, single
#'   dose (default FALSE)
#' @param addOccperDose (\emph{bool}) If TRUE, each dosing will be marked as
#'   different occasion (default FALSE)
#' @param is.timeSinceLastDose (\emph{bool}) If TRUE, time in dataset is
#'   expressed as time after dose (default FALSE).
#' @param dosingOccHeader (\emph{string}) If is.timeSinceLastDose is TRUE,
#'   dosingOccHeader needs to be expressed as header of the column containing
#'   information to distinguish measurements that are after different doses (has
#'   to be in occHeader as well).
#' @param timeFirstDose (\emph{double}) time of the first dose
#' @param doseInterval (\emph{double}) inter-dose interval To specify when
#'   is.multiDose = TRUE
#' @param nbDoses (\emph{int}) number of doses To specify when is.multiDose =
#'   TRUE
#' @param tInf (double) Infusion duration in case of IV infusion (default NULL)
#' @param doseHeader (\emph{string}) Header of the column containing individual
#'   dose amount
#' @param doseAmount (\emph{bool}) Dose amount To specify when doseHeader = NULL
#' @param censTag (\emph{string} or \emph{vector}) Tag(s) representing censored
#'   observation in the dataset. To define when lloq is not NULL (default tags
#'   BLQ, BLOQ, CENS, LOQ, LLOQ).
#' @param lloq (\emph{double}) Lower Limit of quantification (default NULL)
#' @param dataSheet (\emph{string or int}) Sheet to read. Either a string (the
#'   name of a sheet), or an integer (the position of the sheet). If several
#'   data sheets are selected, they are merged. (default first sheet)
#' @param createProject (\emph{bool}) If TRUE, .pkx file will be created
#'   automatically.
#' @param catCov (\emph{vector}) Vector containing categorical covariate header
#'   names.
#' @param contCov (\emph{vector}) Vector containing continuous covariate header
#'   names.
#' @param export (\emph{bool}) To export .xlsx, .csv and .pkx file or not
#' @param outputFilename (\emph{string}) Output file name. Can be absolute or
#'   relative to the folder where the input dataset is stored. If the file
#'   already exists, a new sheet is added to the file (by default the adapted
#'   dataset is saved in a new file with the same extension and a name derived
#'   from the original name) It can be a txt file, a csv file. For excel files,
#'   only xlsx files can be returned. Output file name. Can be absolute or
#'   relative to the folder where the input dataset is stored. It can be a txt,
#'   csv or xlsx file. If an xlsx filename is given, a csv file with the same
#'   name is also outputted.
#'
#' @return a dataframe with data in pkanalix format
#'
#' @importFrom readxl excel_sheets read_excel
#' @importFrom openxlsx loadWorkbook addWorksheet writeData saveWorkbook
#'   write.xlsx createStyle addStyle createWorkbook
#' @importFrom stats median quantile sd
#' @importFrom flextable add_footer_lines add_header_row align
#'   body_add_flextable colformat_double hline vline bold compose flextable
#'   theme_vanilla set_header_labels vline_left padding font as_paragraph
#'   as_chunk as_sub as_sup vline_right
#'
#' @importFrom ggplot2 aes alpha geom_boxplot geom_dotplot geom_hline geom_text
#'   guide_legend element_blank element_line element_rect element_text guides
#'   theme is.ggplot scale_fill_discrete scale_fill_manual scale_x_discrete
#'   stat_summary xlab ylab position_dodge ggplot aes_string facet_wrap
#'
#' @importFrom officer body_add_gg body_remove body_replace_all_text
#'   cursor_backward cursor_reach footers_replace_all_text
#'   headers_replace_all_text fp_border read_docx
#' @export
#' @examples
#' \dontrun{
#'   data <- data.frame(
#'     Id = rep(seq(3), each = 10),
#'     Time = rep(seq(0, 36, 4), 3),
#'     Y = rnorm(10 * 3, mean=10, sd=5)
#'    )
#'    file <- paste0(tempfile(), ".csv")
#'    write.table(x = data, file = file,
#'                row.names = FALSE, quote = FALSE)
#'    formatData(
#'      file, dataType = "plasma",
#'      idHeader = "Id", timeHeader = "Time",
#'      obsHeader = "Y",
#'      is.multiDose = FALSE, timeFirstDose = 0,
#'      is.doseInData = FALSE, doseAmount = 150,
#'    )
#' }
formatData <- function(dataFilename, dataType = "plasma",
                       idHeader = NULL, obsHeader = NULL, timeHeader = NULL,
                       occHeader = NULL, admIdHeader = NULL,
                       urineStartTimeHeader = NULL, urineEndTimeHeader = NULL,
                       urineVolumeHeader = NULL, is.multiDose = FALSE,
                       addOccperDose = FALSE, is.timeSinceLastDose = FALSE,
                       dosingOccHeader= NULL, timeFirstDose = 0,
                       doseInterval = NULL, nbDoses = NULL, tInf = NULL,
                       doseAmount = NULL, doseHeader = NULL, censTag = NULL,
                       lloq = NULL, dataSheet = 1, createProject = F,
                       outputFilename = NULL, catCov = NULL,
                       contCov = NULL, export = TRUE) {

  ##############################################################################
  ## Check Arguments -----------------------------------------------------------
  if (!is.logical(.check_file(dataFilename))) {
    dataFilename <- .check_file(dataFilename)
  }
  if ((!is.logical(is.multiDose)) & (!is.null(is.multiDose))) {
    message("Argument `is.multiDose`, must be logical.")
    return(invisible(FALSE))
  }
  if ((!is.logical(addOccperDose)) & (!is.null(addOccperDose))) {
    message("Argument `addOccperDose`, must be logical.")
    return(invisible(FALSE))
  }
  if ((!is.logical(is.timeSinceLastDose)) & (!is.null(is.timeSinceLastDose))) {
    message("Argument `is.timeSinceLastDose`, must be logical.")
    return(invisible(FALSE))
  }

  is.occasion <- !is.null(occHeader)
  if (!is.element(dataType, c("urine", "plasma"))) {
    message("dataType  must be either `urine` or `plasma`.")
    return(invisible(FALSE))
  }

  # read dataset
  input <- .readDataset(dataFilename, sheet = dataSheet)
  data <- input

  # Check output file
  if (export && !dir.exists(dirname(outputFilename))) {
    message("Creating the output file directory.")
    dir.create(dirname(outputFilename))
  }
  outputFilename <- .check_outputfile(outputFilename,
                                      dataFilename,
                                      sheet = dataSheet)
  if (is.logical(outputFilename)) return(invisible(FALSE))

  if (.getFileExt(outputFilename) %in% c("xls", "xlsx")) {
    if (.getFileExt(outputFilename) == "xls") {
      csvFilename <- sub(x = outputFilename,
                         pattern = ".xls",
                         replacement = ".csv")
    } else {
      csvFilename <- sub(x = outputFilename,
                         pattern = ".xlsx",
                         replacement = ".csv")
    }
  } else {
    csvFilename <- outputFilename
  }

  ##############################################################################
  # Check Headers
  ##############################################################################
  idHeader <- .check_header(data, idHeader, "ID header")
  if (is.logical(idHeader)) return(invisible(FALSE))

  obss <- c()
  for (obs in obsHeader) {
    obss <- c(obss, .check_header(data, obs, "OBSERVATION header"))
    if (is.logical(obs)) return(invisible(FALSE))
  }
  obsHeader <- obss

  if (is.occasion) {
    occs <- c()
    for (occ in occHeader) {
      occs <- c(occs, .check_header(data,
                                    occ,
                                    "OCCASION header"))
      if (is.logical(occ)) return(invisible(FALSE))
    }
    occHeader <- occs
  }

  if (is.timeSinceLastDose) {
    dosingOccHeader = .check_header(data,
                                    dosingOccHeader,
                              "dosingOccHeader header",
                              asTxt = "`is.timeSinceLastDose` is TRUE")
    if (is.logical(dosingOccHeader)) return(invisible(FALSE))
    if(!is.null(dosingOccHeader) && (is.null(occHeader) || !dosingOccHeader %in% occHeader)){
      warning("dosingOccHeader must be in occHeader", call. = F)
    }
  }else{
    if(!is.null(dosingOccHeader)){
      warning("dosingOccHeader is used only if is.timeSinceLastDose=T", call. = F)
    }
  }

  #############################################################################################
  # Managing multiple observation columns
  #############################################################################################
  if (length(obsHeader) > 1) {
    nbObs <- length(obsHeader)
    modifiedDataset <- data.frame(matrix(ncol = ncol(data) + 2, nrow = 0))

    # Rename column OBS if it exists
    if ("OBS" %in% names(data)) {
      colName <- "OBS_"
      while (colName %in% names(data)) {
        colName <- paste0(colName, "_")
      }
      colnames(data)[which(names(data) == "OBS")] <- colName
      message(paste0("Column 'OBS' renamed to '", colName, "'."))
    }

    # Rename column OBS_ID if it exists
    if ("OBS_ID" %in% names(data)) {
      colName <- "OBS_ID_"
      while (colName %in% names(data)) {
        colName <- paste0(colName, "_")
      }
      colnames(data)[which(names(data) == "OBS_ID")] <- colName
      message(paste0("Column 'OBS_ID' renamed to '", colName, "'."))
    }

    # Adding new rows to dataset
    colnames(modifiedDataset) <- c(names(data), "OBS", "OBS_ID")
    for (obs in obsHeader) {
      obsDataset <- data
      obsDataset$OBS <- data[[obs]]
      obsDataset$OBS_ID <- rep(obs, nrow(data))
      modifiedDataset <- rbind(modifiedDataset, obsDataset)
    }
    modifiedDataset <- modifiedDataset[, !names(modifiedDataset) %in% obsHeader]
    data <- modifiedDataset
    obsHeader <- "OBS"
    obsIdHeader <- "OBS_ID"
  }

  #############################################################################################
  # Managing the time in case of plasma and urine
  #############################################################################################
  if (dataType == "plasma") {
    tHeader = .check_header(data,
                            timeHeader,
                            "TIME header",
                            asTxt = "`dataType` is plasma")
    if (is.logical(tHeader)) return(invisible(FALSE))

    if (!is.null(urineStartTimeHeader)) {
      message("When `dataType` is plasma, 'urineStartTimeHeader' is ignored.")
    }
    if (!is.null(urineEndTimeHeader)) {
      message("When `dataType` is plasma, 'urineEndTimeHeader' is ignored.")
    }
    if (!is.null(urineVolumeHeader)) {
      message("When `dataType` is plasma, 'urineVolumeHeader' is ignored.")
    }
  } else if (dataType == "urine") {
    urineStartTimeHeader = .check_header(data,
                                         urineStartTimeHeader,
                                         "Urine START TIME header",
                                         asTxt = "`dataType` is urine")
    urineEndTimeHeader   = .check_header(data,
                                         urineEndTimeHeader,
                                         "Urine END TIME header",
                                         asTxt = "`dataType` is urine")
    urineVolumeHeader    = .check_header(data,
                                         urineVolumeHeader,
                                         "Urine VOLUME header",
                                         asTxt = "`dataType` is urine")

    if (is.logical(urineStartTimeHeader))
      return(invisible(FALSE))

    if (is.logical(urineEndTimeHeader))
      return(invisible(FALSE))

    if (is.logical(urineVolumeHeader))
      return(invisible(FALSE))

    if (!is.null(timeHeader))
      message("When `dataType` is plasma, 'timeHeader'is ignored.")

    tHeader <- urineEndTimeHeader

    # check continuity of time intervals are continuous (no gaps allowed)
    continuity <- by(
      subset(data, select = c(urineStartTimeHeader, urineEndTimeHeader)),
      subset(data, select = subjocc),
      function(df) {
        df <- df[order(df[[urineStartTimeHeader]]),]
        return(all(utils::head(df[[urineEndTimeHeader]], -1) ==
                     utils::tail(df[[urineStartTimeHeader]], -1)))
      }
    )
    if (any(!continuity)) {
      stop("Invalid dataset. In case of urine data, time intervals must be continuous.")
    }
  }

  #############################################################################################
  # Skip lines where time is not a double
  #############################################################################################
  isNA_time = which(is.na(suppressWarnings(as.double(data[[tHeader]]))))
  if (length(isNA_time) > 0) {
    data <- data[-isNA_time,]
  }

  #############################################################################################
  # Check Dose arguments
  #############################################################################################
  timeFirstDose <- .check_double(timeFirstDose, "timeFirstDose")
  if (is.logical(timeFirstDose)) return(invisible(FALSE))

  if (!is.null(doseHeader)) {
    doseHeader = .check_header(data,
                               doseHeader,
                               "Dose amount header")
    if (is.logical(doseHeader)) return(invisible(FALSE))
    is.doseInData = TRUE
    if (!is.null(doseAmount)) {
      message("As dose is defined in dataset ('doseHeader' is provided),
              'doseAmount' is ignored.")
      }
  } else {
    doseAmount <- .check_double(doseAmount,
                                "doseAmount",
                                asTxt = "'doseHeader' is not provided",
                                is.pos = T)
    if (is.logical(doseAmount)) return(invisible(FALSE))
    is.doseInData = FALSE
  }
  if (is.multiDose) {
    doseInterval <- .check_double(doseInterval,
                                  "doseInterval",
                                  asTxt = "`is.multiDose` is TRUE")
    if (is.logical(doseInterval)) return(invisible(FALSE))
    nbDoses <- .check_double(nbDoses,
                             "nbDoses",
                             asTxt = "`is.multiDose` is TRUE",
                             is.int = T)
    if (is.logical(nbDoses)) return(invisible(FALSE))
  } else {
    if (!is.null(doseInterval)) {
      message("When single dose (`is.multiDose` is FALSE),
              'doseInterval' is ignored.")
    }
    if (!is.null(nbDoses)) {
      message("When single dose (`is.multiDose` is FALSE),
              `nbDoses` is ignored.")
    }
  }

  # Check infusion
  if (!is.null(tInf)) {
    tInf <- .check_double(tInf,
                          "tInf",
                          is.pos = T)
    if (is.logical(tInf)) return(invisible(FALSE))
    is.infusion <- TRUE
  } else {
       is.infusion <- FALSE
  }



  #############################################################################################
  # Update time, and OCC in case of multidose
  #############################################################################################
  if (is.multiDose & addOccperDose) {
    # Add the time since last dose
    if (!is.timeSinceLastDose) {
      data$Dosing_OCC = pmin(floor(as.double(data[[tHeader]]) / doseInterval) + 1,
                         nbDoses)
      data$Time_LD = as.double(data[[tHeader]]) -
        (data$Dosing_OCC - 1) * doseInterval
      #occHeader = "Dosing_OCC"
      occHeader = c(occHeader, "Dosing_OCC")
      tHeader = "Time_LD"
      is.multiDose = F
    } else {
      data$Dosing_OCC <- data[[dosingOccHeader]]
      # 3 next lines translates patterns like 'Day 3' into '3', but too specific
      # Day_LD is Dosing_OCC
      #data$Day_LD <- sub(pattern = "DAY", x = data$Day_LD, replacement = "")
      #data$Day_LD <- sub(pattern = "day", x = data$Day_LD, replacement = "")
      #data$Day_LD <- as.double(data$Day_LD )
      #occHeader = "Dosing_OCC"
      occHeader = c(occHeader, "Dosing_OCC")
      is.occasion = T
    }

  }else{
    if (addOccperDose){
      warning("Argument addOccperDose is only used with multiple doses.", call. = FALSE)
    }
  }

  #############################################################################################
  # Check censoring arguments
  #############################################################################################
  censTags <- c("BLQ", "BLOQ", "CENS", "LOQ", "LLOQ")
  is.censored <- FALSE
  if (!is.null(lloq)) {
    censTags <- c(censTags, censTag)
    .check_arg_on_condition(lloq, "lloq")
    is.censored <- TRUE

  }

  # Remove levels of observations
  if (length(levels(data[obsHeader][, ])) > 0) {
    data[obsHeader][,] <- as.character(data[obsHeader][, ])
  }


  # Previous location of code to replace censored and missing observations


  # Add occasions --------------------------------------------------------------
  subjocc <- c(idHeader)
  inputOccHeader <- occHeader
  toccHeader <- c()
  for (occ in occHeader) {
    if (!is.null(occ)) {
      if (sum(is.na(as.double(data[[occ]]))) > 0) {
        # data$OCC <- as.numeric(factor(data[[occHeader]]))
        # occHeader <- "OCC"
        ordered_data <- data[order(data[[idHeader]], data[[occ]], data[[timeHeader]]), ]
        occasions <- c()
        counter <- 1
        for (i in 1:length(data[[idHeader]])) {
          occasions[i] <- counter
          if (i == length(data[[idHeader]])) break
          if (is.na(ordered_data[[occ]][i + 1]) && !is.na(ordered_data[[occ]][i]) ||
              is.na(ordered_data[[occ]][i]) && !is.na(ordered_data[[occ]][i + 1])) {
            counter <- counter + 1
            if (ordered_data[[idHeader]][i] != ordered_data[[idHeader]][i + 1])
              counter <- 0
            next
          }
          if (ordered_data[[occ]][i] != ordered_data[[occ]][i + 1])
            counter <- counter + 1
          if (ordered_data[[idHeader]][i] != ordered_data[[idHeader]][i + 1])
            counter <- 1
        }
        ordered_data$OCC <- occasions
        names(ordered_data)[names(ordered_data) == "OCC"] <- paste0(occ, "_OCC")
        data <- ordered_data
        occ <- paste0(occ, "_OCC")
      }
      subjocc <- c(subjocc, occ)
      toccHeader <- c(toccHeader, occ)
    }
  }
  occHeader <- toccHeader

  ## Add doses -----------------------------------------------------------------
  if (!is.multiDose) {
    nbDoses <- 1
    doseInterval <- 0
  }
  nso <- nrow(unique(subset(data, select = subjocc)))
  if (is.timeSinceLastDose) {
    dosetimes <- rep(timeFirstDose, nso * nbDoses)
  } else {
    dosetimes <- rep(
      seq(timeFirstDose,
          timeFirstDose + (nbDoses - 1) * doseInterval,
          doseInterval),
      nso
    )
  }

  subjoccagg <- sapply(
    stats::aggregate(
      subset(data, select = setdiff(names(data), subjocc)),
      by = as.list(subset(data, select = subjocc)),
      FUN = function(x) length(unique(x))
    ),
    function(x) length(unique(x)) == 1 & x[1] == 1
  )

  fixedCol <- names(subjoccagg[subjoccagg])
  fixedValues <- stats::aggregate(
    data,
    by  = as.list(subset(data, select = subjocc)),
    FUN = function(x) utils::head(x, n = 1)
  )

  if (is.doseInData) {
    if (!doseHeader %in% fixedCol & length(unique(data[[doseHeader]][data[[doseHeader]]!="."]))>1) {
      warning(paste0(
        "Several doses specified per (id, occasion) in column ", doseHeader,
        ". We take the maximum value per (id, occasion)."
      ))
    }
    doseamounts <- as.character(fixedValues[[doseHeader]])
    # Remove all the non-numeric part
    doseamounts <- unlist(regmatches(doseamounts,
                                     gregexpr("[[:digit:]]+\\.*[[:digit:]]*",
                                              doseamounts)))
  } else {
    doseamounts <- rep(doseAmount, nbDoses * nso)
  }
  doseobs <- rep(".", nbDoses * nso)

  if (length(unique(data[[idHeader]])) > 1) {
    doseDf <- do.call("rbind", replicate(
      nbDoses,
      subset(fixedValues, select = c(subjocc, fixedCol)),
      simplify = FALSE)
    )
  } else {
    doseDf <- do.call("rbind", replicate(
      nbDoses,
      subset(fixedValues, select = subjocc),
      simplify = FALSE)
    )
  }


  # if (is.analyte) {
  #   doseDf$ADM_ID <- ifelse(doseDf$AMT == ".", ".", doseDf$OCC)
  # }

  if (is.infusion) {
    doseDf <- cbind(doseDf, list(TINF = rep(tInf, nbDoses * nso)))
    tinfHeaderNewCol <- "TINF"

    # Rename column TINF if it exists
    if ("TINF" %in% names(data)) {
      tinfHeaderNewCol <- "TINF_"
      while (tinfHeaderNewCol %in% names(data)) {
        tinfHeaderNewCol <- paste0(tinfHeaderNewCol, "_")
      }
      doseDf <- .renameColumns(doseDf, c("TINF"), c(tinfHeaderNewCol))
      colnames(doseDf)[ncol(doseDf)] <- tinfHeaderNewCol
      message(paste0("New column 'TINF' renamed to '", tinfHeaderNewCol, "'."))
    }
  }
  if (dataType == "urine") {
    doseDf <- cbind(
      doseDf,
      list(volume = rep(0, nbDoses * nso),
           time1 = dosetimes,
           time2 = dosetimes,
           AMT = doseamounts,
           obs = doseobs
       )
    )
    doseDf <- .renameColumns(
      doseDf,
      c("volume", "time1", "time2", "obs"),
      c(urineVolumeHeader, urineStartTimeHeader, urineEndTimeHeader, obsHeader)
    )
  } else {
    doseDf <- cbind(doseDf, list(time = dosetimes,
                                 AMT = doseamounts,
                                 obs = doseobs))
    doseDf <- .renameColumns(doseDf, c("time", "obs"), c(tHeader, obsHeader))
  }

  # Rename column AMT if it exists
  amtHeaderNewCol <- "AMT"
  if ("AMT" %in% names(data)) {
    amtHeaderNewCol <- "AMT_"
    while (amtHeaderNewCol %in% names(data)) {
      amtHeaderNewCol <- paste0(amtHeaderNewCol, "_")
    }
    message(paste0("New column 'AMT' renamed to '", amtHeaderNewCol, "'."))
  }
  #doseDf <- .renameColumns(doseDf, c("AMT"), c(amtHeaderNewCol))
  colnames(doseDf)[ncol(doseDf)-1] <- amtHeaderNewCol

  data <- merge(data, doseDf, all = TRUE)

  # fill NAN dose columns by "."
  for (h in names(data)) {
    isNA <- is.na(data[h])
    if (length(isNA) > 0) {
      data[h][,] <- as.character(data[h][,])
      data[h][isNA,] <- "."
    }
  }

  # add ADMINISTRATION ID column
  if (!is.null(admIdHeader)) {
    admIdHeaderNewCol <- "ADM_ID"
    if (admIdHeader %in% subjocc | (paste0(admIdHeader, "_OCC") %in% subjocc) ) {
      # Rename column ADM_ID if it exists
      if ("ADM_ID" %in% names(data)) {
        admIdHeaderNewCol <- "ADM_ID_"
        while (admIdHeaderNewCol %in% names(data)) {
          admIdHeaderNewCol <- paste0(admIdHeaderNewCol, "_")
        }
        message(paste0("New column 'ADM_ID' renamed to '", admIdHeaderNewCol, "'."))
      }
    }

    if (admIdHeader %in% subjocc) {
      data[[admIdHeaderNewCol]] <- ifelse(data[[amtHeaderNewCol]] == ".", ".", data[[admIdHeader]])
    } else {
      if (paste0(admIdHeader, "_OCC") %in% subjocc) {
        data[[admIdHeaderNewCol]] <- ifelse(data[[amtHeaderNewCol]] == ".", ".", data[[paste0(admIdHeader, "_OCC")]])
      } else {
        message("'admIdHeader' not in 'occHeader', ADM_ID column won't be added.")
      }
    }
  }


  ## Add Censored observations -------------------------------------------------
  if (is.censored) {

    # Rename column CENS if it exists
    CensHeaderNewCol <- "CENS"
    if ("CENS" %in% names(data)) {
      CensHeaderNewCol <- "CENS_"
      while (CensHeaderNewCol %in% names(data)) {
        CensHeaderNewCol <- paste0(CensHeaderNewCol, "_")
      }
      message(paste0("New column 'CENS' renamed to '", CensHeaderNewCol, "'."))
    }
    data[[CensHeaderNewCol]] <- as.integer(is.element(data[[obsHeader]], censTags[1]))
    for (indexCens in 2:length(censTags)) {
      data[[CensHeaderNewCol]] <- data[[CensHeaderNewCol]] + as.integer(grepl(x = data[[obsHeader]],
                                                pattern = censTags[indexCens]))
    }
    data[[CensHeaderNewCol]][data[[CensHeaderNewCol]]>0] <- 1
    data[obsHeader][data[[CensHeaderNewCol]] == 1,] <- as.character(lloq)
    # remove CENS=0 on dose lines
    data[[CensHeaderNewCol]][data$AMT!="."] <- "."
    # if AMT column was added, move CENS before to match test datasets
    if(!is.null(doseAmount) | !is.null(doseHeader)){
      admIndex <- which(colnames(data)==amtHeaderNewCol)
      data <- data[,c(1:(admIndex-1),ncol(data),admIndex:(ncol(data)-1))]
    }
  }

  ## Replace missing observation -----------------------------------------------
  data <- .replaceMissing(data, obsHeader)


  # stop if data isn't exported
  if (!export) return(data)

  ## Write data ----------------------------------------------------------------
  colOrder <- c(names(input), setdiff(names(data), names(input)))
  # Remove old observation columns from order (if obsHeader was given as a vector)
  if (length(obss) > 1) {
    colOrder <- colOrder[! colOrder %in% obss]
  }

  data <- data[, colOrder]

  if (!is.null(occHeader)) {
    data <- data[order(data[[idHeader]],
                       data[[occHeader[1]]],
                       as.double(data[[tHeader]])),]
  } else {
    data <- data[order(data[[idHeader]], as.double(data[[tHeader]])),]
  }

  if(!exists("obsIdHeader")){
    obsIdHeader <- NULL
  }
  dataColors <- .color_dataset(data,
                               obsHeader,
                               obsIdHeader,
                               tHeader,
                               occHeader,
                               urineVolumeHeader)

  .writeAdaptedData(data, outputFilename, "data", dataFilename, dataColors)
  message("[INFO] Saved pkanalix adapted dataset: ", outputFilename)

  dfName = names(data)
  headerNames <- sub(x = dfName, pattern = '\r', replacement = '_')
  headerNames <- sub(x = headerNames, pattern = '\n', replacement = '_')
  names(data) <- headerNames
  if (outputFilename != csvFilename) {
    utils::write.table(x = data,
                       file = csvFilename,
                       sep = ",",
                       row.names = FALSE,
                       quote = FALSE)
    message("[INFO] Saved pkanalix adapted dataset in csv format: ",
            csvFilename)
  }

  # Write the data into PKanalix
  .lixoftCall("initializeLixoftConnectors", list(software = "pkanalix"))
  headerTypes = rep("ignore", length(dfName))
  headerTypes[which(dfName==idHeader)] = "id"
  headerTypes[which(dfName==tHeader)] = "time"
  headerTypes[which(dfName==obsHeader)] = "observation"
  if (is.censored) {
    headerTypes[which(dfName==CensHeaderNewCol)] = "cens"
  }
  headerTypes[which(dfName==amtHeaderNewCol)] = "amount"
  if (is.infusion) {
    headerTypes[which(dfName==tinfHeaderNewCol)] = "tinf"
  }

  if (!is.null(occHeader)) {
    headerTypes[which(dfName %in% occHeader)] = "occ"
  }
  if (!is.null(doseHeader)) {
    headerTypes[which(dfName==doseHeader)] = "catcov"
  }
  if (!is.null(catCov)) {
    headerTypes[which(dfName %in% catCov)] <- "catcov"
  }
  if (!is.null(contCov)) {
    headerTypes[which(dfName %in% contCov)] <- "contcov"
  }

  pkx_data = list(dataFile = csvFilename,
                  header = names(data),
                  headerTypes = headerTypes)

  # Create project and select settings
  if (createProject) {
    .lixoftCall("newProject", list(data = pkx_data))
    pkx_prj = sub(x = csvFilename, pattern = '.csv$|.txt$|.xls$|.xlsx$', replacement = '.pkx')

    .lixoftCall("setScenario", list(c(nca = T, be = T)))
    .lixoftCall("saveProject", list(pkx_prj))
  }

  return(invisible(FALSE))
}


###################################################################################
# Update the data set headerTypes to add a catCov
# - if not null
# - and if present in the data set
###################################################################################
.addCatCov <- function(catCov = NULL) {

  if (!is.null(catCov)) {
    data = .lixoftCall("getData")
    # Check of the catCov is valid
    for (indexCatCov in 1:length(catCov)) {
      stratCat <- intersect(data$header, catCov[indexCatCov])
      if (length(stratCat) == 0) {
        possibleHeader = data$header[which(data$headerTypes %in% c("ignore", "catcov"))]
        message(paste0("ERROR: Covariate '",  catCov[indexCatCov], "' is not used in the data set. The only available variables are '", paste(possibleHeader, collapse = "', '"),"'."))
        return(invisible(FALSE))
      } else {
        data$headerTypes[which(data$header== catCov[indexCatCov])] <-"catcov"
      }
    }
    .lixoftCall("setData", list(data))
  }
}

.addContCov <- function(contCov = NULL) {

  if (!is.null(contCov)) {
    data = .lixoftCall("getData")
    # Check if the contCov is valid
    for (indexContCov in 1:length(contCov)) {
      stratCat <- intersect(data$header, contCov[indexContCov])
      if (length(stratCat) == 0) {
        possibleHeader = data$header[which(data$headerTypes %in% c("ignore", "contcov"))]
        message(paste0("ERROR: Covariate '",  contCov[indexContCov], "' is not used in the data set. The only available variables are '", paste(possibleHeader, collapse = "', '"),"'."))
        return(invisible(FALSE))
      } else {
        data$headerTypes[which(data$header== contCov[indexContCov])] <-"contcov"
      }
    }
    .lixoftCall("setData", list(data))
  }
}

.skipLines <- function(data, doubleheaders = NULL, naheaders = NULL) {
  # dataset starts with a row with nonnan values
  idxstart <- which(!rowSums(is.na(data)) > 0)[1]
  # dataset starts with a row where headers are double values
  d <- rep(TRUE, nrow(data))
  if (!is.null(doubleheaders)) {
    for (h in doubleheaders) {
      d <- d & suppressWarnings(!is.na(as.double(data[[h]])))
    }
  }
  idxstart <- max(idxstart, which(d)[1])
  data <- data[idxstart:nrow(data),]
  # Remove rows with NAN values on c(id, time, obs)
  if (!is.null(naheaders)) {
    idxna <- which(rowSums(is.na(subset(data, select = naheaders)))==length(naheaders))
    if (length(idxna) > 0) data <- data[-idxna,]
  }
  return(data)
}

.replaceMissing <- function(data, columnNames) {
  for (h in columnNames) {
    replacedStrings <- unique(data[[h]][is.na(suppressWarnings(as.double(data[[h]])))])
    replacedStrings <- replacedStrings[replacedStrings!="."]
    data[[h]][is.na(suppressWarnings(as.double(data[[h]])))] <- "."
    if(length(replacedStrings)>0){
       warning(paste0("Strings in column ",h," have been replaced by '.': ", paste(replacedStrings, collapse = ",")), call. = FALSE)
    }
  }

  return(data)
}

.writeAdaptedData <- function(data, filename, sheet, originalfile, dataColors = NULL) {
  ext <- .getFileExt(filename)
  if (ext == "xlsx") {
    if (file.exists(filename)) {
      if (!is.character(sheet))
        sheet <- paste0("Sheet ", length(excel_sheets(filename)) + 1)
      wb <- loadWorkbook(filename)
    } else {
      wb <- createWorkbook("My name here")
    }
    addWorksheet(wb, sheet)
    # writeData(wb, sheet, data)
    for(i in seq_along(data)){
      writeData(wb, sheet, names(data)[i], startCol = i, startRow = 1)
      icol <- data[[i]]
      for(j in seq_along(icol)){
        x <- icol[[j]]
        if (!suppressWarnings(is.na(as.double(x)))) x <- as.double(x)
        writeData(wb, sheet, x, startCol = i, startRow = j + 1)
      }
    }

    # color table
    if (!is.null(dataColors)) {
      colorsData <- rbind(dataColors$headers, dataColors$data)
      colors <- unique(unlist(colorsData))
      colors <- colors[!is.na(colors)]
      for (color in colors) {
        style <- createStyle(fgFill = color)
        indexes <- as.data.frame(which(colorsData == color, arr.ind = TRUE))
        addStyle(wb, sheet = sheet, style, rows = indexes$row, cols = indexes$col)
      }
    }
    saveWorkbook(wb, filename, overwrite = T)
  } else {
    if (!.getFileExt(originalfile) %in% c("xls", "xlsx")) {
      utils::write.table(
        x = data, file = filename,
        row.names = FALSE, sep = .getDelim(originalfile),
        quote = FALSE
      )
    } else {
      utils::write.table(
        x = data, file = filename, sep = ",",
        row.names = FALSE, quote = FALSE
      )
    }
  }
  return(TRUE)
}

.check_outputfile <- function(outputfile, inputfile, sheet = NULL) {
  if (is.null(outputfile)) {
    ext <- .getFileExt(inputfile)
    if (ext == "xls") ext <- "xlsx"
    outputfile <- .findFileName(inputfile, "pkx", ext = ext)
  } else if (basename(outputfile) == outputfile) {
    outputfile <- file.path(dirname(inputfile), outputfile)
  }
  ext <- .getFileExt(outputfile)
  if (ext == "xls") {
    outputfile <- paste0(tools::file_path_sans_ext(outputfile), ".xlsx")
    warning("xls files are not handled for output format. Adapted dataset will be saved at ",
            outputfile, call. = FALSE)
  }
  if (file.exists(outputfile)) {
    # excel file
    # if (ext == "xlsx") {
    #   if (is.character(sheet) & sheet %in% excel_sheets(outputfile)) {
    #       message("Output file ", outputfile, " already exists and already contains sheet ", sheet, ".")
    #     return(invisible(FALSE))
    #   }
    # } else{
    #   message("Output file ", outputfile, " already exists.")
    #   return(invisible(FALSE))
    # }
    message("Output file ", outputfile, " already exists.")
    return(invisible(FALSE))
  }
  return(outputfile)
}

get_csvfile <- function(file, sheet) {
  file <- tools::file_path_sans_ext(file)
  if (!is.null(sheet) & is.character(sheet)) {
    csvFilename <- .findFileName(file, sheet, ext = "csv")
  } else {
    csvFilename <- .findFileName(file, ext = "csv")
  }
  return(csvFilename)
}

.color_dataset <- function(df, obsHeader, obsIdHeader, timeHeader, occHeader, volHeader) {
  df_color <- data.frame(matrix(ncol = length(df), nrow = nrow(df), dimnames = list(NULL, names(df))))
  names(df_color) <- names(df)
  headers <- as.list(rep(NA, length(df)))
  names(headers) <- names(df)
  if (is.element("CENS", names(df_color))) {
    df_color$CENS[df$CENS == 1] <- "#fab684"
    df_color[[obsHeader]][df$CENS == 1] <- "#fab684"
    df_color$CENS[df$CENS != 1] <- "#fbe7d8"
    headers$CENS <- "#fbe7d8"
  }
  df_color$AMT[df$AMT != "."] <- "#8ad2f0"
  df_color[[timeHeader]][df$AMT != "."] <- "#8ad2f0"
  df_color[[obsHeader]][df$AMT != "."] <- "#8ad2f0"
  df_color$AMT[df$AMT == "."] <- "#c4ebfb"
  headers$AMT <- "#c4ebfb"

  df_color[[obsHeader]][df$AMT == "." & df[[obsHeader]] == "."] <- "#ebeaea"

  #if (is.element("OCC", names(df_color))) {
  #  if (occHeader != "OCC") {
  #    df_color$OCC <- "#ffffc9"
  #    headers$OCC <- "#ffffc9"
  #  }
  #}
  if (!is.null(occHeader)) {
    for(occH in occHeader){
      df_color[[occH]] <- "#ffffc9"
      headers[[occH]] <- "#ffffc9"
    }
  }
  if (!is.null(obsIdHeader)) {
    df_color[[obsIdHeader]] <- "#ffffc9"
    headers[[obsIdHeader]] <- "#ffffc9"
  }
  
  if (!is.null(volHeader)) {
    df_color[volHeader][df$AMT != "."] <- "#76cc76"
  }
  return(list(headers = headers, data = df_color))
}