# ===============================================================================
# Check the names of the parameters
# ===============================================================================
.checkParamNames <- function(parameter) {
  computedParameters <- row.names(.lixoftCall("getNCAIndividualParameters")$information)

  if (!is.null(parameter)) {
    if (!is.vector(parameter)) {
      message("'parameter' must be a vector.")
      return(invisible(FALSE))
    }
    for (index in 1:length(parameter)) {
      if (length(intersect(parameter[index], computedParameters)) == 0) {
        message(paste0("ERROR: Parameter '", parameter[index], "' is not recognized. It was not computed or is not a valid parameter name."))
        return(invisible(FALSE))
      }
    }
  }
  return(parameter)
}

# ===============================================================================
# Check the asMedian parameters to be valid parameters and in the list of
# computed parameters
# ===============================================================================
.checkAsMedianParam <- function(asMedian, param) {
  asMedian <- .checkParamNames(asMedian)
  if (is.logical(asMedian)) {
    return(invisible(FALSE))
  }
  # Check if asMedian are in the computed params
  if (length(asMedian) > 0) {
    for (indexMed in 1:length(asMedian)) {
      if (length(intersect(param, asMedian[indexMed])) == 0) {
        message(paste0("ERROR: parameter `", asMedian[indexMed], "` is not in the list of computed parameters."))
        return(invisible(FALSE))
      }
    }
  }

  return(asMedian)
}


# ===============================================================================
# Check the names of the stats
# ===============================================================================
.checkStatNames <- function(stats) {
  availableStats <- c(
    "Mean", "Min", "Q1", "Median", "Q3", "Max", "SD", "SE",
    "CV", "CV%", "GeoMean", "GeoSD", "GeoCV", "GeoCV%", "N",
    "hmean"
  )

  if (!is.null(stats)) {
    if (!is.vector(stats)) {
      message("ERROR: 'stats' must be a vector.")
      return(invisible(FALSE))
    }
    for (index in 1:length(stats)) {
      if (length(intersect(tolower(stats[index]), tolower(availableStats))) == 0) {
        message(paste0("ERROR: Stat '", stats[index], "' is not recognized.
                       Available stats are ", paste(availableStats, collapse = ", "), "."))
        return(invisible(FALSE))
      }
    }
  }
  return(stats)
}

# ===============================================================================
# Check the name of the covariate and the associated modalities
# - cov is a string corresponding to the covariate name
# - modality is vector of strings correspondings to the modalities
#   associated to the covariate cov
# ===============================================================================
.checkCOVName <- function(covName, modality = NULL) {
  availableCov <- .lixoftCall("getCovariateInformation")$name
  dataDescription <- .lixoftCall("getData")
  availableOcc <- dataDescription$header[dataDescription$headerTypes == "occ"]
  if (!is.null(covName)) {
    isCovariate <- length(intersect(covName, availableCov)) != 0
    isOccasion <- length(intersect(covName, availableOcc)) != 0
    if (!isCovariate && !isOccasion) {
      message <- paste0(
        "ERROR: Covariate or occasion ",
        covName,
        " is not used in the data set."
      )
      if (length(availableCov) > 0) {
        message <- paste0(
          message,
          " Available covariates: ",
          paste(availableCov, collapse = ", "),
          "."
        )
      }
      if (length(availableOcc) > 0) {
        message <- paste0(
          message,
          " Available occasions: ",
          paste(availableOcc, collapse = ", "),
          "."
        )
      }
      message(message)
      return(invisible(FALSE))
    }
    if (!is.null(modality) && isCovariate) {
      modCov <- unique(as.character(.lixoftCall("getCovariateInformation")$covariate[[covName]]))
      for (index in 1:length(modality)) {
        if (length(intersect(modality[index], modCov)) == 0) {
          message(paste0("ERROR: Modality ", modality[index], " is not a modality of ", covName, ". Available modalities of ", covName, " are ", paste(modCov, collapse = ", "), "."))
          return(invisible(FALSE))
        }
      }
    }
    if (!is.null(modality) && isOccasion) {
      obsName <- .lixoftCall("getObservationInformation")$name
      modCov <- unique(as.character(.lixoftCall("getObservationInformation")[[obsName]][[covName]]))
      for (index in 1:length(modality)) {
        if (length(intersect(modality[index], modCov)) == 0) {
          message(paste0("ERROR: Modality ", modality[index], " is not a modality of ", covName, ". Available modalities of ", covName, " are ", paste(modCov, collapse = ", "), "."))
          return(invisible(FALSE))
        }
      }
    }
  }
  return(covName)
}

# ===============================================================================
# Check that there is no filterinf nor renaming with a null covariate
# ===============================================================================
.checkCovAvailability <- function(covName, renameCov, filterCov, strCovName, strRenameCov, strFilterCov) {
  bIsOK <- T
  if (is.null(covName)) {
    if (!is.null(renameCov)) {
      message(paste0("ERROR: Renaming action '", strRenameCov, "' will not be performed as '", strCovName, "' is NULL."))
      bIsOK <- F
    }
    if (!is.null(filterCov)) {
      message(paste0("ERROR: Filtering action '", strFilterCov, "' will not be performed as '", strCovName, "' is NULL."))
      bIsOK <- F
    }
  }
  return(bIsOK)
}

.checkAddline <- function(addLine, addText, addLineColor) {
  bIsOK <- T
  if (is.null(addLine)) {
    if (!is.null(addText)) {
      message(paste0("Warning: Adding text 'addText' will not be performed as 'addLine' is NULL."))
      bIsOK <- F
    }
    if (!is.null(addLineColor)) {
      message(paste0("Warning: Choosing color action 'addLineColor' will not be performed as 'addLine' is NULL."))
      bIsOK <- F
    }
  }
  return(bIsOK)
}


# ===============================================================================
# Get the number of modalities of a category / 0 if NULL
# ===============================================================================
.getNbModalities <- function(catName) {
  if (!is.null(catName)) {
    nbCat <- length(as.character(unique(.lixoftCall("getCovariateInformation")$covariate[[catName]])))
  } else {
    nbCat <- 0
  }
  return(nbCat)
}
