#' Plot mean observed data
#'
#' \code{getPLOT_MEANOBS} returns mean observed data plot as a ggplot2 object.
#'
#' @param xLab (\emph{string}) x-axis label.
#' @param yLab (\emph{string}) y-axis label.
#' @param bLog (\emph{bool} or \emph{"both"}) If TRUE, y-axis will be in in log
#'   scale, if "both", both axes will be in log scale. FALSE by default.
#' @param sameYlimit (\emph{bool}) If TRUE and splitCat is defined, all plots
#'   will have the same y-axis limits.
#' @param bCtrough (\emph{bool}) If TRUE, only Ctrough values will be plotted.
#' @param colorCat (\emph{string}) Categorical covariate to color stratify by.
#' @param renameCat (\emph{vector}) Displayed names of colorCat modalities.
#' @param filterCat (\emph{vector}) colorCat modalities to filter out.
#' @param colorCat2 (\emph{string}) Categorical covariate to color stratify by.
#' @param renameCat2 (\emph{vector}) Displayed names of colorCat modalities.
#' @param filterCat2 (\emph{vector}) colorCat modalities to filter out.
#' @param splitCat (\emph{string}) Categorical covariate to split plots by.
#' @param renameSplitCat (\emph{vector}) Displayed names of splitCat modalities.
#' @param filterSplitCat (\emph{vector}) splitCat modalities to filter out.
#' @param addLine (\emph{double} or \emph{list}) Value of y to display as a
#'   horizontal line (if double, it will be shown on all plots in case of
#'   splits, otherwise need to specify modalities).
#' @param addText (\emph{string} or \emph{list}) Text to show above the
#'   horizontal line, needs to have the same format as addLine.
#' @param addLineColor (\emph{string} or \emph{list}) Horizontal line color,
#'   needs to have the same format as addLine.
#' @param binsSettings (\emph{string} or \emph{list}) a list of settings for
#'   time axis binning for observation statistics computation. More details in
#'   lixoftConnectors plotObservedData documentation.
#'
#' @return A ggplot2 object.
#'
#' @examples
#' \dontrun{
#' library(lixoftConnectors)
#'
#' initializeLixoftConnectors("pkanalix")
#' demoPath <- getDemoPath()
#' loadProject(paste0(demoPath, "/3.bioequivalence/project_crossover_bioequivalence.pkx"))
#' setNCASettings(
#'   partialAucTime = list(T, c(0, 24)),
#'   computedNCAParameters = c("AUCINF_obs", "Cmax", "Tmax", "Tlast")
#' )
#' runNCAEstimation()
#'
#' meanObs <- getPLOT_MEANOBS(
#'   splitCat = "FORM",
#'   renameSplitCat = c("ref" = "Reference", "test" = "Test"),
#'   colorCat = "SEQ"
#' )
#' meanObs
#' }
#'
#' @export
getPLOT_MEANOBS <- function(xLab = "Time", yLab = "Plasma concentration", bLog = F, sameYlimit = T, bCtrough = F,
                             colorCat = NULL, renameCat = NULL, filterCat = NULL,
                             colorCat2 = NULL, renameCat2 = NULL, filterCat2 = NULL,
                             splitCat = NULL, renameSplitCat = NULL, filterSplitCat = NULL,
                             addLine = NULL, addText = NULL, addLineColor = NULL, binsSettings = NULL){

  #-----------------------------------------------------------------------
  # Check the inputs
  #-----------------------------------------------------------------------
  if (!((is.logical(bLog) || (bLog =="both")))) {
    message(paste0("ERROR:'bLog' should be a boolean."))
    return(invisible(FALSE))
  }
  if (!is.logical(sameYlimit)) {
    message(paste0("ERROR:'sameYlimit' should be a boolean."))
    return(invisible(FALSE))
  }

  #-----------------------------------------------------------------------
  # Initialize the settings
  #-----------------------------------------------------------------------
  pSettings <- list(dots=F, lines=F, mean=T, error=T,
                    errorMethod="standardDeviation", legend=T, scales= "free_x")

  if(sameYlimit){
    pSettings$scales= "fixed"
  }else{
    pSettings$scales= "free_y"
  }
  if(!is.null(xLab)){pSettings$xlab <- .addUnits(paramStr = xLab, paramName = "Tmax")}
  if(!is.null(yLab)){pSettings$ylab <- .addUnits(paramStr = yLab, paramName = "Cmax")}

  #-----------------------------------------------------------------------
  # Check the covariates
  #-----------------------------------------------------------------------
  if(is.logical(.checkCOVName(colorCat))){return(invisible(FALSE))}
  if(!.checkCovAvailability(colorCat, renameCat, filterCat, 'colorCat', 'renameCat', 'filterCat')){return(invisible(FALSE))}
  if(is.logical(.checkCOVName(colorCat2))){return(invisible(FALSE))}
  if(!.checkCovAvailability(colorCat2, renameCat2, filterCat2, 'colorCat2', 'renameCat2', 'filterCat2')){return(invisible(FALSE))}
  if(is.logical(.checkCOVName(splitCat))){return(invisible(FALSE))}
  if(!.checkCovAvailability(splitCat, renameSplitCat, filterSplitCat, 'splitCat', 'renameSplitCat', 'filterSplitCat')){return(invisible(FALSE))}

  #-----------------------------------------------------------------------
  # Update the categories modalities to take renaming into account
  #-----------------------------------------------------------------------
  hCat <- .updateModalities(colorCat, renameCat, filterCat)
  if(is.logical(hCat)){return(invisible(FALSE))}
  modCat <- hCat[[1]];displayCat <- hCat[[2]]; nbCat = length(modCat)

  hCat2 <- .updateModalities(colorCat2, renameCat2, filterCat2)
  if(is.logical(hCat2)){return(invisible(FALSE))}
  modCat2 <- hCat2[[1]];displayCat2 <- hCat2[[2]];nbCat2 = length(modCat2)

  sCat <- .updateModalities(splitCat, renameSplitCat, filterSplitCat)
  if(is.logical(sCat)){return(invisible(FALSE))}
  modSplitCat <- sCat[[1]];displaySplitCat <- sCat[[2]]

  #-----------------------------------------------------------------------
  # Define all the groups
  #-----------------------------------------------------------------------
  colorGrp <-  filterGrp <- splitGrp <- NULL
  if((nbCat>1)&(nbCat2==1)){
    colorGrp = list(name = colorCat)
  }else if((nbCat2>1)&(nbCat==1)){
    colorGrp = list(name = colorCat2)
  }else if((nbCat>1)&(nbCat2>1)){
    colorGrp = list(list(name = colorCat), list(name = colorCat2))
  }

  #--- Manage the filtering group
  filterGrp <- .getFilterGrp(colorCat, filterCat, colorCat2, filterCat2, splitCat, filterSplitCat)

  #--- Manage the Split group
  if(!is.null(splitCat)){
    splitGrp = list(name = splitCat)
  }

  data <- .lixoftCall("getChartsData", list(plotName = "plotObservedData", colorGroup = colorGrp, filter = filterGrp, splitGroup = splitGrp, computeSettings = list(binsSettings = binsSettings)))

  if (bCtrough) {
    dCtrough <- data$statistics
    # Pick the trough concentration
    dCtrough <- dCtrough[dCtrough$time == 0, ]
    days <- dCtrough$color
    if (nbCat2 == 1) {
      # The category with the days is colorCat
      days <- sub(days, pattern = paste0(colorCat2,": "), replacement = "")
      pSettings$xlab = colorCat
    } else {
      # The category with the days is colorCat2
      if (nbCat > 1) {
        for (indexCat in 1:nbCat) {
          days<- sub(days, pattern = paste0(colorCat, ": ", modCat[indexCat], '.'), replacement = "")
        }
      }
      days <- sub(days, pattern = paste0(colorCat2,": "), replacement = "")
      pSettings$xlab = colorCat2
    }
    # days <- sub(x = tolower(days), pattern = "day", replacement = "")
    # days <- as.double(days)
    dCtrough$time = days

    dCtrough$color <- as.character(dCtrough$color)
    # Remove the info in the color
    if (nbCat2 == 1) {
      # The category with the days is colorCat
      dCtrough$color = "All"
    } else {
      # The category with the days is colorCat2
      if (nbCat > 1) {
        for (indexCat in 1:nbCat) {
          pattern = paste0(colorCat,": ",modCat[indexCat])
          for(index in 1:length(dCtrough$color)){
            if(length(grep(x = dCtrough$color[index], pattern = pattern))>0){dCtrough$color[index] <- pattern}
          }
        }
      }
    }
    data$statistics <- dCtrough
  }

  if(!is.null(colorCat2)){
    data <- .renameData(data, type = "statistics", colName = "color", colorCat, renameCat, refactor = F)
    data <- .renameData(data, type = "statistics", colName = "color", colorCat2, renameCat2, refactor = F)
    data$statistics$color <- factor(data$statistics$color)
  }else{
    data <- .renameData(data, type = "statistics", colName = "color", colorCat, renameCat)
  }
  data <- .renameData(data, type = "observations", colName = "split", splitCat, renameSplitCat)
  data <- .renameData(data, type = "statistics", colName = "split", splitCat, renameSplitCat)

  if(!is.null(splitCat)&is.null(renameSplitCat)){
    renameSplit <-  modSplitCat
    names(renameSplit) <- modSplitCat
    data <- .renameData(data, type = "observations", colName = "split", splitCat, renameSplit)
    data <- .renameData(data, type = "statistics", colName = "split", splitCat, renameSplit)
   }

  ddf <- data

  if(nbCat+nbCat2>2){
    pSettings$legend = T
  }else{
    pSettings$legend = F
  }
  pSettings$meanMethod = "arithmetic";pSettings$ylog = F;
  pLin <- .lixoftCall("plotObservedData", list(data = data, settings=pSettings)) + ggPlotTheme()

  pSettings$meanMethod = "geometric";pSettings$ylog = T;
  pLog <- .lixoftCall("plotObservedData", list(data = data, settings=pSettings)) + ggPlotTheme()

  .checkAddline(addLine, addText, addLineColor)

  if(!pSettings$legend){
    pLin <- pLin + theme(legend.position = "none")
    pLog <- pLog + theme(legend.position = "none")
  }

  pLin <- .addLine(pLin, addLine, addText, addLineColor, splitCat, renameSplitCat)
  pLog <- .addLine(pLog, addLine, addText, addLineColor, splitCat, renameSplitCat)

  if(bLog=="both"){
    pLin <- pLin + theme(legend.position = "bottom", legend.justification = "center", legend.direction = "horizontal", legend.title = element_blank())+
        guides(colour = guide_legend(title.position = "left"))
    pLog <- pLog + theme(legend.position = "bottom", legend.justification = "center", legend.direction = "horizontal", legend.title = element_blank())+
        guides(colour = guide_legend(title.position = "left"))
    if(pSettings$legend){
      pp <- ggpubr::ggarrange(pLin,pLog, common.legend = T, legend='bottom', ncol = 2)
    }else{
      pp <- ggpubr::ggarrange(pLin,pLog, common.legend = T, legend='none', ncol = 2)
    }
  }else if(bLog){
    pp <- pLog
  }else{
    pp <- pLin
  }

  return(pp)
}

#' Get mean observed data as ggplot2 object
#'
#' @param xLab (\emph{string}) x-axis label.
#' @param yLab (\emph{string}) y-axis label.
#' @param bLog (\emph{bool} or \emph{"both"}) y-axis in log scale or both axes
#'   in log scale.
#' @param sameYlimit (\emph{bool}) If TRUE, all plots will have the same y-axis
#'   limit.
#' @param filterCat (\emph{vector}) Categories to filter.
#' @param filterMod (\emph{vector}) Categories to filter.
#' @param splitCat (\emph{string}) Categorical covariate to split plots by.
#' @param renameSplitCat (\emph{vector}) Display names of categories.
#' @param filterSplitCat (\emph{vector}) Categories to filter.
#' @param addLine (\emph{double} or \emph{vector}) Value of y to display as
#'   horizontal line.
#' @param addText (\emph{string} or \emph{vector}) Text to show above horizontal
#'   line.
#' @param addLineColor (\emph{string} or \emph{vector}) Line color (e.g. "red",
#'   c("yellow", "red")).
#' @param merge (\emph{bool}) If TRUE, all plots will be merged to one (default
#'   TRUE).
#'
#' @return a ggplot2 object
#'
#' @examples
#' \dontrun{
#' library(lixoftConnectors)
#'
#' initializeLixoftConnectors("pkanalix")
#' demoPath <- getDemoPath()
#' loadProject(paste0(demoPath, "/3.bioequivalence/project_crossover_bioequivalence.pkx"))
#' setNCASettings(
#'   partialAucTime = list(T, c(0, 24)),
#'   computedNCAParameters = c("AUCINF_obs", "Cmax", "Tmax", "Tlast")
#' )
#' runNCAEstimation()
#'
#' indivObs <- getPLOT_INDIVOBS(
#'   splitCat = "FORM",
#'   renameSplitCat = c("ref" = "Reference", "test" = "Test")
#' )
#' indivObs
#' }
#'
#' @export
getPLOT_INDIVOBS <- function(xLab = "Time", yLab = "Plasma concentration", bLog = F, sameYlimit = T,
                              filterCat = NULL, filterMod = NULL,
                              splitCat = NULL, renameSplitCat = NULL, filterSplitCat = NULL,
                              addLine = NULL, addText = NULL, addLineColor = NULL, merge = T){

  if (!merge) {
    return(.getPLOT_INDIVOBSs(xLab, yLab, bLog, sameYlimit, filterCat, filterMod,
                              splitCat, renameSplitCat, filterSplitCat, addLine, addText, addLineColor))
  }

  #-----------------------------------------------------------------------
  # Check the inputs
  #-----------------------------------------------------------------------
  if (!((is.logical(bLog) || (bLog =="both")))) {
    message(paste0("ERROR:'bLog' should be a boolean."))
    return(invisible(FALSE))
  }
  if (!is.logical(sameYlimit)) {
    message(paste0("ERROR:'sameYlimit' should be a boolean."))
    return(invisible(FALSE))
  }

  #-----------------------------------------------------------------------------
  # Initialize the settings
  #-----------------------------------------------------------------------------
  pSettings <- list(legend=T, dots=T, lines=T, mean=F, error=F, units=F, grid=T, ylog = bLog,
                    scales= "free_x")
  if(!is.null(xLab)){pSettings$xlab <- .addUnits(paramStr = xLab, paramName = "Tmax")}
  if(!is.null(yLab)){pSettings$ylab <- .addUnits(paramStr = yLab, paramName = "Cmax")}
  IDname = .lixoftCall("getData")$header[which(.lixoftCall("getData")$headerTypes=="id")]

  if (sameYlimit) {
    pSettings$scales= "fixed"
  } else {
    pSettings$scales= "free_y"
  }

  #-----------------------------------------------------------------------
  # Check the covariates
  #-----------------------------------------------------------------------
  if (is.logical(.checkCOVName(splitCat)))
    return(invisible(FALSE))

  if (!.checkCovAvailability(splitCat, renameSplitCat, filterSplitCat,
                             'splitCat', 'renameSplitCat', 'filterSplitCat'))
    return(invisible(FALSE))

  hSplit<- .updateModalities(splitCat, renameSplitCat, filterSplitCat)
  if (is.logical(hSplit))
    return(invisible(FALSE))

  modSplit <- hSplit[[1]]
  displaySplit <- hSplit[[2]]
  nbSlit = length(modSplit)

  filterCat <- .checkCOVName(filterCat, filterMod)
  if (is.logical(filterCat))
    return(invisible(FALSE))

  if (is.null(filterCat)) {
    if (!is.null(filterCat)) {
      message(paste0("ERROR: Filtering action 'filterMod' will not be performed as 'filterCat' is NULL."))
      return(invisible(FALSE))
    }
  }

  #-----------------------------------------------------------------------------
  # Define all the groups
  #-----------------------------------------------------------------------------
  splitGrp <-  filterGrp <- NULL
  if (!is.null(splitCat)) {
    splitGrp = list(list(name = splitCat))
  }
  filterGrp <- .getFilterGrp(filterCat, filterMod, splitCat, filterSplitCat)

  data <- .lixoftCall("getChartsData", list(plotName = "plotObservedData",
                                            colorGroup = list(name = IDname),
                                            splitGroup = splitGrp,
                                            filter = filterGrp))

  # Rename the legend for the IDs to remove "IDname: " before display
  data$observations$color <- as.character(data$observations$color)
  data$observations$color <- sub(x = data$observations$color,
                                 pattern = paste0(IDname,": "),
                                 replacement = "")
  data$observations$color <- factor(data$observations$color,
                                    levels = unique(data$observations$color))

  # Rename the legend for the categorical covariate before display
  data <- .renameData(data, type = "observations",
                      colName = "split", splitCat, renameSplitCat)

  pSettings$ylog = F;
  pLin <- .lixoftCall("plotObservedData", list(data = data, settings = pSettings)) + ggPlotTheme() +
    theme(legend.position = "bottom", legend.justification = "center", legend.direction = "horizontal", legend.title = element_blank())+
    guides(colour = guide_legend(title.position = "left"))

  pSettings$ylog = T;
  pLog <- .lixoftCall("plotObservedData", list(data = data, settings=pSettings)) + ggPlotTheme() +
    theme(legend.position = "bottom", legend.justification = "center", legend.direction = "horizontal", legend.title = element_blank())+
    guides(colour = guide_legend(title.position = "left"))

  pLin <- .addLine(pLin, addLine, addText, addLineColor, splitCat, renameSplitCat)
  pLog <- .addLine(pLog, addLine, addText, addLineColor, splitCat, renameSplitCat)

  if (bLog == "both") {
    pp <- ggpubr::ggarrange(pLin,pLog, common.legend = T, legend='bottom', ncol = 2)
  } else if (bLog) {
    pp <- pLog
  } else {
    pp <- pLin
  }

  return(pp)
}

.getPLOT_INDIVOBSs <- function(xLab = "Time", yLab = "Plasma concentration", bLog = F, sameYlimit = T,
                              filterCat = NULL, filterMod = NULL,
                              splitCat = NULL, renameSplitCat = NULL, filterSplitCat = NULL,
                              addLine = NULL, addText = NULL, addLineColor = NULL){


  #-----------------------------------------------------------------------
  # Check the inputs
  #-----------------------------------------------------------------------
  if(!((is.logical(bLog)||(bLog =="both")))){message(paste0("ERROR:'bLog' should be a boolean."));return(invisible(FALSE))}
  if(!is.logical(sameYlimit)){message(paste0("ERROR:'sameYlimit' should be a boolean."));return(invisible(FALSE))}

  #-----------------------------------------------------------------------
  # Check the covariates
  #-----------------------------------------------------------------------
  if(is.logical(.checkCOVName(splitCat))){return(invisible(FALSE))}
  if(!.checkCovAvailability(splitCat, renameSplitCat, filterSplitCat, 'splitCat', 'renameSplitCat', 'filterSplitCat')){return(invisible(FALSE))}
  hSplit<- .updateModalities(splitCat, renameSplitCat, filterSplitCat)
  if(is.logical(hSplit)){return(invisible(FALSE))}
  modSplit <- hSplit[[1]];displaySplit <- hSplit[[2]];nbSplit = length(modSplit)

  filterCat <- .checkCOVName(filterCat, filterMod)
  if(is.logical(filterCat)){return(invisible(FALSE))}
  if(is.null(filterCat)){
    if(!is.null(filterCat)){message(paste0("ERROR: Filtering action 'filterMod' will not be performed as 'filterCat' is NULL."));return(invisible(FALSE))}
  }

  if(nbSplit==1){
    p <- getPLOT_INDIVOBS(xLab, yLab, bLog, sameYlimit,
                             filterCat, filterMod, splitCat, renameSplitCat, filterSplitCat, addLine, addText, addLineColor)

    if(bLog!="both"){
      p <-  ggpubr::ggarrange(p, common.legend = T, legend='bottom', ncol = 1)
    }
    if(!is.null(filterMod)){
      leg <- paste0(filterMod)
    }else{
      leg <- ""
    }
    p <- ggpubr::annotate_figure(p,leg)

  }else{
    str = "list("
    for(index in 1:nbSplit){
      p <- getPLOT_INDIVOBS(xLab, yLab, bLog, sameYlimit,
                             filterCat, filterMod, splitCat, renameSplitCat[index], filterSplitCat = modSplit[index],
                             addLine, addText, addLineColor)

      if(!is.null(filterMod)){
        leg <- paste0(filterMod, " - ")
      }else{
        leg <- ""
      }
      leg <- paste0(leg, gsub(x = displaySplit[index], pattern = '\n', replacement = ': '))
      if(bLog!="both"){
        p <-  ggpubr::ggarrange(p, common.legend = T, legend='bottom', ncol = 1)
      }
      p <- ggpubr::annotate_figure(p,leg)
      eval(parse(text = paste0("p_",index," <- p")))
      str = paste0(str,"p_",index,',')
    }
    str = paste0(substr(str, 1, nchar(str)-1),")")
    pOut <- eval(parse(text = str))
    p <- pOut
  }
  return(p)

}

#' Get plot by indiv/using individual fits as ggplot2 object
#'
#' @param xLab (\emph{string}) x-axis label.
#' @param yLab (\emph{string}) y-axis label.
#' @param bLog (\emph{bool} or \emph{"both"}) y-axis in log scale or both axes in log scale.
#' @param lambda_z (\emph{bool}) Show lambda_z.
#' @param splitCat (\emph{string}) Categorical covariate to split plots by.
#' @param renameSplitCat (\emph{vector}) Display names of categories.
#' @param filterSplitCat (\emph{vector}) Categories to filter.
#' @param colorCat (\emph{string}) Covariate to color by.
#' @param filterCat (\emph{string}) Category to filter by.
#' @param filterMod (\emph{string}) Category to filter by.
#'
#' @return a ggplot2 object
#'
#' @examples
#' \dontrun{
#' library(lixoftConnectors)
#'
#' initializeLixoftConnectors("pkanalix")
#' demoPath <- getDemoPath()
#' loadProject(paste0(demoPath, "/3.bioequivalence/project_crossover_bioequivalence.pkx"))
#' setNCASettings(
#'   partialAucTime = list(T, c(0, 24)),
#'   computedNCAParameters = c("AUCINF_obs", "Cmax", "Tmax", "Tlast")
#' )
#' runNCAEstimation()
#'
#' indivObs <- getPLOT_byINDIV(
#'   splitCat = "FORM",
#'   renameSplitCat = c("ref" = "Reference", "test" = "Test")
#' )
#' indivObs
#' }
#'
#' @export
getPLOT_byINDIV <- function(xLab = "Time", yLab = "Plasma concentration", bLog = F, lambda_z = F,
                             splitCat = NULL, renameSplitCat = NULL, filterSplitCat = NULL,
                             colorCat = NULL, filterCat = NULL, filterMod = NULL){

  #-----------------------------------------------------------------------
  # Initialize the settings
  #-----------------------------------------------------------------------
  pSettings <- list(legend=T, obsLines=T, units=F, grid=F, ylog = bLog, lambda_z=lambda_z,
                    splitOccasions = F)
  if(!is.null(xLab)){pSettings$xlab <- .addUnits(paramStr = xLab, paramName = "Tmax")}
  if(!is.null(yLab)){pSettings$ylab <- .addUnits(paramStr = yLab, paramName = "Cmax")}
  IDname = .lixoftCall("getData")$header[which(.lixoftCall("getData")$headerTypes=="id")]

  if(lambda_z){
    pSettings$obsLines = F
    pSettings$ylog = T
  }
  #-----------------------------------------------------------------------
  # Check the covariates
  #-----------------------------------------------------------------------
  colorCat <- .checkCOVName(colorCat)

  splitCat <- .checkCOVName(splitCat)
  if(is.logical(splitCat)){return(invisible(FALSE))}
  if(is.null(splitCat)){
    if(!is.null(renameSplitCat)){message(paste0("ERROR: Renaming action 'renameSplitCat' will not be performed as 'splitCat' is NULL."));return(invisible(FALSE))}
    if(!is.null(filterSplitCat)){message(paste0("ERROR: Filtering action 'filterSplitCat' will not be performed as 'splitCat' is NULL."));return(invisible(FALSE))}
  }
  hSplit<- .updateModalities(splitCat, renameSplitCat, filterSplitCat)
  if(is.logical(hSplit)){return(invisible(FALSE))}
  modSplit <- hSplit[[1]];displaySPlit <- hSplit[[2]]
  nbSlit = length(modSplit)

  filterCat <- .checkCOVName(filterCat, filterMod)
  if(is.logical(filterCat)){return(invisible(FALSE))}
  if(is.null(filterCat)){
    if(!is.null(filterCat)){message(paste0("ERROR: Filtering action 'filterMod' will not be performed as 'filterCat' is NULL."));return(invisible(FALSE))}
  }

  #-----------------------------------------------------------------------
  # Define all the groups
  #-----------------------------------------------------------------------
  splitGrp <- colorGrp <- filterGrp <- NULL

  if(!is.null(splitCat)){
    pSettings$splitOccasions = T
  }

  if(!is.null(colorCat)){
    colorGrp = list(name=colorCat)
  }

  filterGrp <- .getFilterGrp(filterCat, filterMod, splitCat, filterSplitCat)

  data <- .lixoftCall("getChartsData", list(plotName = "plotNCAIndividualFits",
                                            colorGroup = colorGrp,
                                            splitGroup = splitGrp,
                                            filter = filterGrp))

  #-----------------------------------------------------------------------
  pp <- .lixoftCall("plotNCAIndividualFits", list(data = data,
                                                  settings=pSettings)) + ggPlotTheme()

  pp <- pp + theme(legend.position = "bottom", legend.justification = "center", legend.direction = "horizontal", legend.title = element_blank())
  pp <- pp + guides(colour = guide_legend(title.position = "left"))
  return(pp)
}

#' Get boxplot of a variable as a ggplot2 object
#'
#' @param param (\emph{string}) NCA parameter to get box plot of.
#' @param colorCat (\emph{string}) Categorical covariate to color by.
#' @param renameCat (\emph{vector}) Display names of categories.
#' @param filterCat (\emph{vector}) Categories to filter.
#' @param splitCat (\emph{string}) Categorical covariate to split plots by.
#' @param renameSplitCat (\emph{vector}) Display names of categories.
#' @param filterSplitCat (\emph{vector}) Categories to filter.
#' @param splitCat2 (\emph{string}) Categorical covariate to split plots by.
#' @param renameSplitCat2 (\emph{vector}) Display names of categories.
#' @param filterSplitCat2 (\emph{vector}) Categories to filter.
#' @param addPoint (\emph{bool}) If TRUE, points are shown on the plot
#' @param sameYlimit (\emph{bool}) If TRUE, all plots will have the same Y axis scale.
#' @param colorMean (\emph{string}) Color of the mean point.
#' @param addLine (\emph{double} or \emph{vector}) Value of y to display as horizontal line.
#' @param addText (\emph{string} or \emph{vector}) Text to show above horizontal line.
#' @param addLineColor (\emph{string} or \emph{vector}) Line color (e.g. "red", c("yellow", "red")).
#' @param merge (\emph{bool}) If TRUE, one object will be returned (default TRUE).
#' @param ncol (\emph{int}) Number of columns, if merge = FALSE.
#'
#' @return a ggplot2 object
#'
#' @examples
#' \dontrun{
#' library(lixoftConnectors)
#'
#' initializeLixoftConnectors("pkanalix")
#' demoPath <- getDemoPath()
#' loadProject(paste0(demoPath, "/3.bioequivalence/project_crossover_bioequivalence.pkx"))
#' setNCASettings(
#'   partialAucTime = list(T, c(0, 24)),
#'   computedNCAParameters = c("AUCINF_obs", "Cmax", "Tmax", "Tlast")
#' )
#' runNCAEstimation()
#'
#' boxPlot <- getBOXPLOT(
#'   param = "Cmax",
#'   splitCat = "FORM",
#'   renameSplitCat = c("ref" = "Reference", "test" = "Test"),
#'   addPoint = F
#' )
#' boxPlot
#' }
#'
#' @export
getBOXPLOT <- function(param,
                        colorCat = NULL, renameCat = NULL, filterCat = NULL,
                        splitCat = NULL, renameSplitCat = NULL, filterSplitCat = NULL,
                        splitCat2 = NULL, renameSplitCat2 = NULL, filterSplitCat2 = NULL,
                        addPoint = T, sameYlimit = T, colorMean = "black",
                        addLine = NULL, addText = NULL, addLineColor = NULL, merge = T, ncol = NULL){

  if (!merge) {
    return(.getBOXPLOTs(param, ncol, colorCat, renameCat, filterCat, splitCat,
                        renameSplitCat, filterSplitCat, splitCat2, renameSplitCat2,
                        filterSplitCat2, addPoint, sameYlimit, colorMean,
                        addLine, addText, addLineColor))
  }
  #-----------------------------------------------------------------------
  # Get the initial datas and check the params
  #-----------------------------------------------------------------------
  if(!is.logical(addPoint)){  message(paste0("ERROR: 'addPoint' should be a boolean."));return(invisible(FALSE))}
  if(!is.logical(sameYlimit)){  message(paste0("ERROR: 'sameYlimit' should be a boolean."));return(invisible(FALSE))}

  #-----------------------------------------------------------------------
  # Get the initial datas and check the params
  #-----------------------------------------------------------------------
  if(is.logical(.checkParamNames(param))){return(invisible(FALSE))}
  indivParam <- .lixoftCall("getNCAIndividualParameters", list(param))$parameters

  #-----------------------------------------------------------------------
  # Check the covariates
  #-----------------------------------------------------------------------
  if(is.logical(.checkCOVName(colorCat))){return(invisible(FALSE))}
  if(!.checkCovAvailability(colorCat, renameCat, filterCat, 'colorCat', 'renameCat', 'filterCat')){return(invisible(FALSE))}
  if(is.logical(.checkCOVName(splitCat))){return(invisible(FALSE))}
  if(!.checkCovAvailability(splitCat, renameSplitCat, filterSplitCat, 'splitCat', 'renameSplitCat', 'filterSplitCat')){return(invisible(FALSE))}
  if(is.logical(.checkCOVName(splitCat2))){return(invisible(FALSE))}
  if(!.checkCovAvailability(splitCat2, renameSplitCat2, filterSplitCat2, 'splitCat2', 'renameSplitCat2', 'filterSplitCat2')){return(invisible(FALSE))}

  #-----------------------------------------------------------------------
  # Update the categories modalities to take renaming into account
  #-----------------------------------------------------------------------
  hCat <- .updateModalities(colorCat, renameCat, filterCat)
  if(is.logical(hCat)){return(invisible(FALSE))}
  indivParam <- .updateIndivParam(indivParam, colorCat, renameCat, filterCat)
  nbCat = length(hCat[[1]])

  hSplitCat <- .updateModalities(splitCat, renameSplitCat, filterSplitCat)
  if(is.logical(hSplitCat)){return(invisible(FALSE))}
  indivParam <- .updateIndivParam(indivParam, splitCat, renameSplitCat, filterSplitCat)
  nbSplitCat = length(hSplitCat[[1]])

  hSplitCat2 <- .updateModalities(splitCat2, renameSplitCat2, filterSplitCat2)
  if(is.logical(hSplitCat2)){return(invisible(FALSE))}
  indivParam <- .updateIndivParam(indivParam, splitCat2, renameSplitCat2, filterSplitCat2)
  nbSplitCat2 = length(hSplitCat2[[2]])
  modSplit2 <- hSplitCat2[[1]];displaySplit2 <- hSplitCat2[[2]];
  if(nbCat==1){
    colorCat = factor(0)
  }

  #-----------------------------------------------------------------------------
  # Generate the plot
  #-----------------------------------------------------------------------------
  if(nbSplitCat2>1){
    if(nbSplitCat>1){
      if (nbCat == 1) {
        eval(parse(text = paste0("p <- ggplot(indivParam, aes(x=reorder(",splitCat,", suppressWarnings(as.numeric(", splitCat, "))), y=",param,", fill = ''))")))
      } else {
        eval(parse(text = paste0("p <- ggplot(indivParam, aes(x=reorder(",splitCat,", suppressWarnings(as.numeric(", splitCat, "))), y=",param,", fill = ",colorCat,"))")))
      }
    }else{
      eval(parse(text = paste0("p <- ggplot(indivParam, aes(x=reorder(",colorCat,", suppressWarnings(as.numeric(", colorCat, "))), y=",param,", fill = ",colorCat,"))")))
    }
    if(sameYlimit){
      eval(parse(text = paste0('p <- p + facet_wrap(. ~ ',splitCat2,')+theme(strip.background =element_rect(fill="white"))')))
    }else{
      eval(parse(text = paste0('p <- p + facet_wrap(. ~ ',splitCat2,',  scale="free")+theme(strip.background =element_rect(fill="white"))')))
      addPoint <- FALSE
    }
  }else{
    if(nbSplitCat>1){
      if (nbCat == 1) {
        eval(parse(text = paste0("p <- ggplot(indivParam, aes(x=reorder(",splitCat,", suppressWarnings(as.numeric(", splitCat, "))), y=",param,", fill = ''))")))
      } else {
        eval(parse(text = paste0("p <- ggplot(indivParam, aes(x=reorder(",splitCat,", suppressWarnings(as.numeric(", splitCat, "))), y=",param,", fill = ",colorCat,"))")))
      }
    }else{
      eval(parse(text = paste0("p <- ggplot(indivParam, aes(x=reorder(",colorCat,", suppressWarnings(as.numeric(", colorCat, "))), y=",param,", fill = ",colorCat,"))")))
    }
  }
  p <- p + xlab("") + geom_boxplot(position=position_dodge(0.8), alpha = .5)
  if(colorCat!=0){
    p <- p + scale_x_discrete(limits = levels(indivParam[[colorCat]]))
  }
  if(addPoint){
    bw <- (max(indivParam[[param]])-min(indivParam[[param]]))/30
    # if(nbSplitCat2>1){
    #   for(indexCat2 in 1:nbSplitCat2){
    #     vect = indivParam[[param]][which(indivParam[[splitCat2]]==hSplitCat2[[2]][indexCat2])]
    #     bw = min(bw, (max(vect)-min(vect))/30)
    #   }
    # }
    p <- p + geom_dotplot(binaxis='y', stackdir='center',
                          binwidth = bw,
                          position = position_dodge(0.8), dotsize = 1)
  }
  p <- p + stat_summary(fun=mean, geom = "point", shape=15, color = colorMean, position=position_dodge(0.8), size = 4, show.legend=F)


  if(nbSplitCat > 1){
    p <- suppressMessages(p  + scale_fill_discrete(name = splitCat, labels = levels(indivParam[[splitCat]])))
    p <- suppressMessages(p + scale_fill_manual(values=.colorTemplate(nbSplitCat)))
  }else{
    if(nbCat > 1){
      p <- suppressMessages(p  + scale_fill_discrete(name = colorCat, labels = levels(indivParam[[colorCat]])))
    }
  }
  if(nbCat>1){
    p <- suppressMessages(p + scale_fill_manual(values=.colorTemplate(nbCat)))
  }
  yLab <- eval(parse(text = paste0("expression(",.addUnits(paramStr =  .renameParamNamesPlots(param), paramName = param, isExp=T),")")))

  # Add lines
  .checkAddline(addLine, addText, addLineColor)

  p <- .addLine(p, addLine, addText, addLineColor, splitCat2, renameSplitCat2)

  p <- p + ylab(yLab) + ggPlotTheme()
  if(nbSplitCat+nbSplitCat2>2){
    if(nbCat>1){
      p <- p + theme(legend.key = element_rect(colour = NA, fill = NA),
                     legend.position = "bottom", legend.justification = "center", legend.direction = "horizontal", legend.title = element_blank())
    }else{
      p <- p + theme(legend.position = "none")
    }
  }else{
    p <- p + theme(legend.position = "none")
  }

  # if(colorCat==0){
  #   p <- p +  theme(axis.ticks.x = element_blank(),axis.text.x = element_blank())
  # }

  return(suppressWarnings(p))
}

.getBOXPLOTs <- function(params, ncol=NULL,
                        colorCat = NULL, renameCat = NULL, filterCat = NULL,
                        splitCat = NULL, renameSplitCat = NULL, filterSplitCat = NULL,
                        splitCat2 = NULL, renameSplitCat2 = NULL, filterSplitCat2 = NULL,
                        addPoint = T, sameYlimit = T, colorMean = "black",
                        addLine = NULL, addText = NULL, addLineColor = NULL){

  #-----------------------------------------------------------------------
  # Get the initial datas and check the params
  #-----------------------------------------------------------------------
  if(is.logical(.checkParamNames(params))){return(invisible(FALSE))}
  nbParam = length(params)

  if(is.null(ncol)){
    ncol = nbParam
  }

  #-----------------------------------------------------------------------
  # Check the covariates
  #-----------------------------------------------------------------------
  if(is.logical(.checkCOVName(colorCat))){return(invisible(FALSE))}
  if(!.checkCovAvailability(colorCat, renameCat, filterCat, 'colorCat', 'renameCat', 'filterCat')){return(invisible(FALSE))}
  if(is.logical(.checkCOVName(splitCat))){return(invisible(FALSE))}
  if(!.checkCovAvailability(splitCat, renameSplitCat, filterSplitCat, 'splitCat', 'renameSplitCat', 'filterSplitCat')){return(invisible(FALSE))}
  if(is.logical(.checkCOVName(splitCat2))){return(invisible(FALSE))}
  if(!.checkCovAvailability(splitCat2, renameSplitCat2, filterSplitCat2, 'splitCat2', 'renameSplitCat2', 'filterSplitCat2')){return(invisible(FALSE))}

  #-----------------------------------------------------------------------
  # Update the categories modalities to take renaming into account
  #-----------------------------------------------------------------------
  hCat <- .updateModalities(colorCat, renameCat, filterCat)
  if(is.logical(hCat)){return(invisible(FALSE))}
  nbCat = length(hCat[[1]])

  hSplitCat <- .updateModalities(splitCat, renameSplitCat, filterSplitCat)
  if(is.logical(hSplitCat)){return(invisible(FALSE))}
  nbSplitCat = length(hSplitCat[[1]])

  hSplitCat2 <- .updateModalities(splitCat2, renameSplitCat2, filterSplitCat2)
  if(is.logical(hSplitCat2)){return(invisible(FALSE))}
  nbSplitCat2 = length(hSplitCat2[[2]])

  str = NULL
  for(index in 1:nbParam){
    p <- getBOXPLOT(params[index],
                     colorCat, renameCat, filterCat,
                     splitCat, renameSplitCat , filterSplitCat ,
                     splitCat2, renameSplitCat2 , filterSplitCat2 ,
                     addPoint, sameYlimit, colorMean,addLine , addText , addLineColor)
    eval(parse(text = paste0("p_",params[index]," <- p")))
    str = paste0(str,"p_",params[index],',')
  }

  if((nbSplitCat+nbSplitCat2>2)&(nbCat>1)){
    pp <- eval(parse(text = paste0("ggpubr::ggarrange(",substr(str, 1, nchar(str)-1),", common.legend = TRUE, legend='bottom', ncol = ",ncol,")")))
  }else{
    pp <- eval(parse(text = paste0("ggpubr::ggarrange(",substr(str, 1, nchar(str)-1),", common.legend = TRUE, legend='none', ncol = ",ncol,")")))

  }

  return(pp)
}


################################################################################
# Function to rename the splitCat in the data for the plots
################################################################################
.renameData <- function(data, type = "observations", colName, catName, renameCat, refactor = T){

  # Rename the legend for the categorical covariate before display
  # Do the work on the observations
  dataObs <- data[[type]]
  dataObs[[colName]] <- as.character( dataObs[[colName]] )
  dataObs[[colName]] <- sub(x =  dataObs[[colName]] , pattern = paste0(catName,": "), replacement = "")

  if(!is.null(renameCat)){
    xx = dataObs[[colName]]
    orderCatIncr <- sort(renameCat)
    orderCatDecr <- sort(renameCat, decreasing = T)

    for(index in 1:length(renameCat)){
      if(refactor){
        indCat <- grep(x = dataObs[[colName]], pattern = names(orderCatIncr)[index])
        xx[indCat] <- orderCatIncr[index]
      }else{
        xx <- sub(xx, pattern = names(orderCatDecr)[index], replacement = orderCatDecr[index])
      }
    }
    data[[type]][[colName]] <- xx

    if(refactor){
      data[[type]][[colName]] <- factor(data[[type]][[colName]], levels=as.vector(renameCat))
    }
  } else {
    data[[type]][[colName]] <- dataObs[[colName]]
  }
  return(data)
}

################################################################################
# Rename the modalities of a data frame
################################################################################
.updateIndivParam <- function(data, catName = NULL, renameCat = NULL,
                              filterCat = NULL){

  if(!is.null(catName)){
    modCat <- as.character(unique(data[[catName]]))
    nbCat <- length(modCat)

    # Filter the data set to take the filterCat modalities
    if(!is.null(filterCat)){
      for(indexFilt in 1 : length(filterCat)){
        if(length(intersect(modCat, filterCat[indexFilt])) == 0){
          message(paste0("ERROR: The modality '",filterCat[indexFilt] ,"', asked to be filtered, is not a modality of '",renameCat,"'."))
          return(invisible(FALSE))
        }
      }
      iFilter <-  which(data[[catName]] %in% unique(filterCat))
      modCat <- modCat[modCat %in% unique(filterCat)]
      nbCat <- length(modCat)
      data <- data[iFilter,]
      data[[catName]] <- factor(data[[catName]], levels=as.vector(filterCat))
    }

    # Rename the categories associated for the filtered data
    if(!is.null(renameCat)){
      for(indexFilt in 1 : length(names(renameCat))){
        if(length(intersect(modCat, names(renameCat)[indexFilt])) == 0){
          message(paste0("ERROR: The modality '",names(renameCat)[indexFilt] ,"', asked to be renamed, is not (anymore) a modality of '",renameCat,"'."))
          return(invisible(FALSE))
        }
      }
      if(length(intersect(modCat, names(renameCat))) < nbCat){
        message(paste0("ERROR: The renaming of categories should be a vector with ",nbCat, " values, i.e. the same as the number of modalities of ",renameCat," after filtering."))
        return(invisible(FALSE))
      }else{
        displayCat = as.vector(renameCat)
        modCat <- names(renameCat)
        data[[catName]] <- as.character(data[[catName]])
        xx = data[[catName]]
        for(indexFilt in 1 : length(modCat)){
          xx[which(data[[catName]] %in% modCat[indexFilt])] <- displayCat[indexFilt]
        }
        data[[catName]] <- xx
        data[[catName]] <- factor(data[[catName]], levels=as.vector(renameCat))
      }
    }
  }

  return(data)
}

################################################################################
# Add the text and the line to the plot
################################################################################
addTextPlot <- function(p, addLine=NULL, addText = NULL, addLineColor= NULL, splitName = NULL) {

  if (is.null(addLineColor)) {
    addLineColor = rep("black", length(addLine))
  }
  if (length(addLineColor) < length(addLine)) {
    addLineColor = rep(addLineColor[1], length(addLine))
  }

  if (is.null(addText)) {
    addText = rep("", length(addLine))
  }
  if (length(addText) < length(addLine)) {
    addText = rep(addText[1], length(addLine))
  }
  if (!is.null(addLine)) {
    for (index in 1:length(addLine)) {
      # Add a text and line
      if (is.null(splitName)) {
        p <- p + geom_hline(yintercept = as.double(addLine[index]), color = addLineColor[index], linetype = "dashed", size=1)
        p <- p + geom_text(y = as.double(addLine[index]), x = Inf,label = paste0(addText[index],"   "), hjust = 1, vjust = -.5)

      } else {
        p <- p + geom_hline(data=data.frame(yint=as.double(addLine[index]), split = splitName), aes_string(yintercept = "yint"), color = addLineColor[index], linetype = "dashed", size=1)
        p <- p + geom_text(data=data.frame(yint=as.double(addLine[index]), split = splitName), aes_string(y = "yint"), x = Inf,label = paste0(addText[index],"   "), hjust = 1, vjust = -.5)
      }
    }
  }
  return(p)
}

################################################################################
# Get the filter group for plots with 3 filter possibilities
################################################################################
.getFilterGrp <- function(colorCat = NULL, filterCat = NULL,
                          colorCat2 = NULL, filterCat2 = NULL,
                          splitCat = NULL, filterSplitCat = NULL){

  #--- Manage the filtering group
  if(!is.null(filterCat)){
    filterGrpCat = list(name = colorCat, cat = filterCat);bFiltCat = 1
  }else{
    bFiltCat = 0
  }
  if(!is.null(filterCat2)){
    filterGrpCat2 = list(name = colorCat2, cat = filterCat2);bFiltCat2 = 1
  }else{
    bFiltCat2 = 0
  }
  if(!is.null(filterSplitCat)){
    filterGrpSplit = list(name = splitCat, cat = filterSplitCat);bFiltSplit = 1
  }else{
    bFiltSplit = 0
  }

  if((bFiltCat+bFiltCat2+bFiltSplit)==1){
    if(bFiltCat==1){
      filterGrp = filterGrpCat
    }else if(bFiltCat2==1){
      filterGrp = filterGrpCat2
    }else{
      filterGrp = filterGrpSplit
    }
  }else if((bFiltCat+bFiltCat2+bFiltSplit)==2){
    if(bFiltCat==0){
      filterGrp = list(filterGrpCat2, filterGrpSplit)
    }else if(bFiltCat2==0){
      filterGrp = list(filterGrpCat, filterGrpSplit)
    }else{
      filterGrp = list(filterGrpCat, filterGrpCat2)
    }
  }else if((bFiltCat+bFiltCat2+bFiltSplit)==3){
    filterGrp = list(filterGrpCat, filterGrpCat2, filterGrpSplit)
  }else{
    filterGrp <- NULL
  }
  return(filterGrp)
}

.addLine <- function(plot, addLine = NULL, addText = NULL, addLineColor = NULL,
                     splitCat = NULL, renameSplitCat = NULL) {
  if (is.null(names(addLine))) {
    plot <- addTextPlot(plot, addLine, addText, addLineColor, splitName = NULL)
  } else {
    if (is.null(splitCat)) {
      message("WARNING: splitCat is not defined, lines won't be plotted.")
    } else {
      if (!is.null(renameSplitCat)) {
        for (splitMod in names(addLine)) {
          if (splitMod %in% names(renameSplitCat)) {
            plot <- addTextPlot(plot, addLine[[splitMod]], addText[[splitMod]], addLineColor[[splitMod]], splitName = renameSplitCat[splitMod])
          }
        }
      } else {
        for (splitMod in names(addLine)) {
          plot <- addTextPlot(plot, addLine[[splitMod]], addText[[splitMod]], addLineColor[[splitMod]], splitName = splitMod)
        }
      }
    }
  }
  return(plot)
}
