#' Summary table of NCA metrics
#'
#' \code{getNCASummaryTable} returns the summary table of NCA metrics as a
#' Flextable object.
#'
#' @param paramListSummary (\emph{vector}) NCA metrics to be displayed in the
#'   table. All available options can be found in the table
#'   \href{https://pkanalix.lixoft.com/nca-parameters/}{here}.
#' @param asMedian (\emph{vector}) NCA metrics for which the calculated summary
#'   statistics is median.
#' @param stat (\emph{vector}) Summary statistics to compute. Available options
#'   are min, q1, median, q3, max, mean, sd, se, cv, cv%, geomean, geosd, geocv,
#'   geocv%, hmean and n.
#' @param nbSignDigit (\emph{int}) Number of significant digits used to describe
#'   the data.
#' @param horizCatName (\emph{string}) Name of the categorical covariate to
#'   split the data by horizontally.
#' @param renameHorizCat (\emph{vector}) Displayed names for each of the
#'   horizCatName modalities.
#' @param filterHorizCat (\emph{vector}) Names of horizCatName modalities not to
#'   be displayed in the table.
#' @param horizCat2Name (\emph{string}) Name of the categorical covariate to
#'   split the data by horizontally in the second level.
#' @param renameHorizCat2 (\emph{vector}) Displayed names for each of the
#'   horizCat2Name modalities.
#' @param filterHorizCat2 (\emph{vector}) Names of horizCat2Name modalities not
#'   to be displayed in the table.
#' @param vertCatName (\emph{string}) Name of the categorical covariate to split
#'   the data by vertically.
#' @param renameVertCat (\emph{vector}) Displayed names for each of the
#'   vertCatName modalities.
#' @param filterVertCat (\emph{vector}) Names of vertCatName modalities not to
#'   be displayed in the table.
#' @param displayNbIndiv (\emph{bool}) Display number of profiles used in the
#'   calculation in the header.
#' @param filterBy (\emph{vector}) Vector containing one or more of the
#'   following strings: "Flag_AUC_%Extrap_pred", "Flag_Rsq_adjusted",
#'   "Flag_Span". Note: included acceptance options need to be set in PKanalix
#'   project in order for filtering to function properly.
#' @param filterID (\emph{vector}) Vector containing IDs of individuals to
#'   filter out of the analysis.
#' @param doseNormalizationUnit (\emph{string}) String containing normalization
#'   unit, if dose was normalized.
#'
#' @return A Flextable object.
#'
#' @examples
#' \dontrun{
#' library(lixoftConnectors)
#'
#' initializeLixoftConnectors("pkanalix")
#' demoPath <- getDemoPath()
#' loadProject(paste0(demoPath, "/3.bioequivalence/project_crossover_bioequivalence.pkx"))
#' setNCASettings(
#'   partialAucTime = list(T, c(0, 24)),
#'   computedNCAParameters = c("AUCINF_obs", "Cmax", "Tmax", "Tlast")
#' )
#' runNCAEstimation()
#'
#' NCASumTable <- getNCASummaryTable(
#'   paramListSummary = c("AUCINF_obs", "Cmax", "AUC_0_24", "Tmax", "Tlast"),
#'   asMedian = c("Tmax", "Tlast"),
#'   stat = c("Mean", "CV%", "HMean"),
#'   nbSignDigit = 4,
#'   horizCatName = "FORM",
#'   renameHorizCat = c("ref" = "Reference", "test" = "Test"),
#'   vertCatName = "SEQ",
#'   filterID = c("1", "2")
#' )
#' NCASumTable
#' }
#'
#' @export
getNCASummaryTable <- function(paramListSummary = NULL, asMedian = NULL, stat = c("Mean", "CV%"), nbSignDigit = 3,
                               horizCatName = NULL, renameHorizCat = NULL, filterHorizCat = NULL,
                               horizCat2Name = NULL, renameHorizCat2 = NULL, filterHorizCat2 = NULL,
                               vertCatName = NULL, renameVertCat = NULL, filterVertCat = NULL,
                               displayNbIndiv = T, filterBy = NULL, filterID = NULL, doseNormalizationUnit = NULL) {

  #-----------------------------------------------------------------------------
  # Check the parameters
  #-----------------------------------------------------------------------------
  if (is.null(paramListSummary)) {
    param <- row.names(.lixoftCall("getNCAIndividualParameters")$information)
  } else {
    param <- .checkParamNames(paramListSummary)
    if (is.logical(param)) {
      return(invisible(FALSE))
    }
    param <- sort(param)
  }
  indivParam <- .lixoftCall("getNCAIndividualParameters", list(param))$parameters

  # Filter indivParam by acceptance criteria
  if (length(filterBy) > 0) {
    for (criteria in filterBy) {
      if (!(criteria %in% names(indivParam))) {
        message(paste0("WARNING: data won't be filtered by '", criteria, "', since the parameter of that name doesn't exist."))
        next
      }
      indivParam <- indivParam[indivParam[criteria] == 1, ]
    }
  }

  # Filter indivParam by ID
  if (length(filterID) > 0) {
    indivParam <- indivParam[!(indivParam$ID %in% filterID), ]
  }

  # Add transformed covariate if needed
  indivParam <- .addTransformedCov(indivParam)
  nbParam <- length(param)

  #-----------------------------------------------------------------------------
  # Manage the units which can be missing in the projects
  #-----------------------------------------------------------------------------
  unitsLine <- as.character(.lixoftCall("getNCAIndividualParameters", list(param))$information$units)
  if (length(unitsLine) == 0) {
    message("Warning: Units were not used in the PKanalix project.")
    indivParamSummary <- data.frame(Variable = param)
    bUnits <- 0
  } else {
    indivParamSummary <- data.frame(Variable = param, Units = as.character(.lixoftCall("getNCAIndividualParameters", list(param))$information$units))
    bUnits <- 1
    if (!is.null(doseNormalizationUnit)) {
      for (i in 1:length(param)) {
        indivParamSummary$Units[i] <- paste0(
          indivParamSummary$Units[i],
          .addNormalization(
            param[i],
            doseNormalizationUnit
          )
        )
      }
    }
  }

  #-----------------------------------------------------------------------------
  # Check that the parameters is asMedian are part of the computed datas
  #-----------------------------------------------------------------------------
  if (is.logical(.checkAsMedianParam(asMedian, param))) {
    return(invisible(FALSE))
  }

  #-----------------------------------------------------------------------
  # Check the stats
  #-----------------------------------------------------------------------
  if (is.logical(.checkStatNames(stat))) {
    return(invisible(FALSE))
  }

  nbStat <- length(stat)
  if (nbStat == 0) {
    message("ERROR: 'stat' can not be NULL.")
    return(invisible(FALSE))
  }
  if ("median" %in% stat) {
    message("WARNING: asMedian argument will be ignored as median is in stat.")
    asMedian <- NULL
    stat <- tolower(stat)
  }

  #-----------------------------------------------------------------------
  # Check nbSignDigit, displayNbIndiv
  #-----------------------------------------------------------------------
  if (!(is.numeric(nbSignDigit) & (nbSignDigit %% 1 == 0))) {
    message("ERROR: 'nbSignDigit' must be an integer.")
    return(invisible(FALSE))
  }
  if (!is.logical(displayNbIndiv)) {
    message("ERROR: 'displayNbIndiv' must be an boolean.")
    return(invisible(FALSE))
  }

  #-----------------------------------------------------------------------------
  # Get the check of the categories
  #-----------------------------------------------------------------------------
  if (is.logical(.checkCOVName(horizCatName))) {
    return(invisible(FALSE))
  }
  nbHorizCatOriginal <- .getNbModalities(horizCatName)
  if (!.checkCovAvailability(
    horizCatName, renameHorizCat, filterHorizCat,
    "horizCatName", "renameHorizCat", "filterHorizCat"
  )) {
    return(invisible(FALSE))
  }

  if (is.logical(.checkCOVName(horizCat2Name))) {
    return(invisible(FALSE))
  }
  nbHorizCat2Original <- .getNbModalities(horizCat2Name)
  if (!.checkCovAvailability(
    horizCat2Name, renameHorizCat2, filterHorizCat2,
    "horizCat2Name", "renameHorizCat2", "filterHorizCat2"
  )) {
    return(invisible(FALSE))
  }

  if (is.logical(.checkCOVName(vertCatName))) {
    return(invisible(FALSE))
  }
  nbVertCatOriginal <- .getNbModalities(vertCatName)
  if (!.checkCovAvailability(
    vertCatName, renameVertCat, filterVertCat,
    "vertCatName", "renameVertCat", "filterVertCat"
  )) {
    return(invisible(FALSE))
  }

  #-----------------------------------------------------------------------------
  # Update the categories modalities to take renaming into account
  #-----------------------------------------------------------------------------
  nbOcc <- sum(.lixoftCall("getData")$headerTypes == "occ")

  hCat <- .updateModalities(horizCatName, renameHorizCat, filterHorizCat)
  if (is.logical(hCat)) {
    return(invisible(FALSE))
  }
  modHorizCat <- hCat[[1]]
  displayHorizCat <- hCat[[2]]
  nbHorizCat <- length(modHorizCat)

  hCat2 <- .updateModalities(horizCat2Name, renameHorizCat2, filterHorizCat2)
  if (is.logical(hCat2)) {
    return(invisible(FALSE))
  }
  modHorizCat2 <- hCat2[[1]]
  displayHorizCat2 <- hCat2[[2]]
  nbHorizCat2 <- length(modHorizCat2)


  vCat <- .updateModalities(vertCatName, renameVertCat, filterVertCat)
  if (is.logical(vCat)) {
    return(invisible(FALSE))
  }
  modVertCat <- vCat[[1]]
  displayVertCat <- vCat[[2]]
  nbVertCat <- length(modVertCat)

  if (nbHorizCat > 1 && nbHorizCat2 > 1) {
    displayHorizCat <- rep(displayHorizCat, nbHorizCat2)
  }

  # Update the categories name to put the number of each individual in the category
  if (displayNbIndiv & (nbVertCat < 2)) {
    if (nbHorizCat == 1) {
      if (nbHorizCat2 == 1) {
        displayHorizCat <- paste0(displayHorizCat, " (n=", nrow(unique(indivParam[, 1:(1 + nbOcc)])), ")")
      } else {
        for (indexHorizCat2 in 1:nbHorizCat2) {
          IDs <- indivParam[which(indivParam[[horizCat2Name]] == modHorizCat2[indexHorizCat2]), 1:(1 + nbOcc)]
          displayHorizCat2[indexHorizCat2] <- paste0(displayHorizCat2[indexHorizCat2], "\n (n=", nrow(unique(IDs)), ")")
        }
      }
    } else {
      if (nbHorizCat2 == 1) {
        for (indexHorizCat in 1:nbHorizCat) {
          IDs <- indivParam[which(indivParam[[horizCatName]] == modHorizCat[indexHorizCat]), 1:(1 + nbOcc)]
          if (is.null(nrow(unique(IDs)))) {
            displayHorizCat[indexHorizCat] <- paste0(displayHorizCat[indexHorizCat], "\n (n=", length(IDs), ")")
          } else {
            displayHorizCat[indexHorizCat] <- paste0(displayHorizCat[indexHorizCat], "\n (n=", nrow(unique(IDs)), ")")
          }
        }
      } else {
        for (indexHorizCat in 1:nbHorizCat) {
          for (indexHorizCat2 in 1:nbHorizCat2) {
            IDs <- indivParam[indivParam[[horizCatName]] == modHorizCat[indexHorizCat] & indivParam[[horizCat2Name]] == modHorizCat2[indexHorizCat2], 1:(1 + nbOcc)]
            N <- nrow(unique(IDs))
            displayHorizCat[(indexHorizCat2 - 1) * nbHorizCat + indexHorizCat] <- paste0(displayHorizCat[(indexHorizCat2 - 1) * nbHorizCat + indexHorizCat], "\n (n=", N, ")")
          }
        }
      }
    }
  }

  #-----------------------------------------------------------------------------
  # Initialize the indivParamSummary
  #-----------------------------------------------------------------------------
  idOffset <- 0
  if (nbVertCat > 1) {
    for (indexVert in 2:nbVertCat) {
      indivParamSummary <- rbind(
        indivParamSummary,
        data.frame(Variable = param, Units = as.character(.lixoftCall("getNCAIndividualParameters", list(param))$information$units))
      )
    }
  }

  #-----------------------------------------------------------------------------
  # Compute the stats by modalities of vertCatName
  #-----------------------------------------------------------------------------
  for (indexHorz2 in 1:nbHorizCat2) {
    for (indexHorz1 in 1:nbHorizCat) {
      df_Summary <- NULL
      for (indexVert in 1:nbVertCat) {
        # get the index (ID#OCC) of the needed
        indexFilter <- 1:length(indivParam$ID)
        if (nbHorizCatOriginal > 1) {
          colVal <- indivParam[[horizCatName]][indexFilter]
          indexFilter <- indexFilter[which(colVal == modHorizCat[indexHorz1])]
        }
        if (nbHorizCat2Original > 1) {
          colVal <- indivParam[[horizCat2Name]][indexFilter]
          indexFilter <- indexFilter[which(colVal == modHorizCat2[indexHorz2])]
        }
        if (nbVertCatOriginal > 1) {
          colVal <- indivParam[[vertCatName]][indexFilter]
          indexFilter <- indexFilter[which(colVal == modVertCat[indexVert])]
        }
        dataFilt <- indivParam[indexFilter, (1 + nbOcc) + (1:nbParam)]
        if (nbParam == 1) {
          dataFilt <- data.frame(dataFilt)
        }
        df_Summary <- rbind(
          df_Summary,
          .getStatSummary(data = dataFilt, stat, asMedian, indexOffset = idOffset)
        )
      }
      idOffset <- idOffset + nbStat
      indivParamSummary <- cbind(indivParamSummary, df_Summary)
    }
  }

  # Add number of individuals to vertical split
  if (displayNbIndiv && nbVertCat > 1 && nbHorizCat == 1 && nbHorizCat2 == 1) {
    for (indexVert in 1:nbVertCat) {
      IDs <- indivParam[which(indivParam[[vertCatName]] == modVertCat[indexVert]), 1:(1 + nbOcc)]
      displayVertCat[indexVert] <- paste0(displayVertCat[indexVert], "\n (n=", nrow(unique(IDs)), ")")
    }
  }

  # If there is a vertical split, add the categories on the left
  if (nbVertCat > 1) {
    df_split <- NULL
    for (indexVert in 1:nbVertCat) {
      df_split <- c(df_split, displayVertCat[indexVert], rep("", nbParam - 1))
    }
    df_split <- data.frame(df_split)
    names(df_split) <- vertCatName
    indivParamSummary <- cbind(df_split, indivParamSummary)
  }

  # Update the labels to "remove" the fake names to be well perceived by flextable
  labelNames <- list()
  for (index in 1:length(names(indivParamSummary))) {
    nameCol <- names(indivParamSummary)[index]
    labelNames[index] <- strsplit(nameCol, split = "_", fixed = T)[[1]][1]
  }
  names(labelNames) <- names(indivParamSummary)

  # Apply the rounding to the summary
  indivParamSummary <- .getRound(indivParamSummary, nbSignDigit)

  # removing empty columns
  # tempSummary <- indivParamSummary
  # for (i in 1:ncol(indivParamSummary)) {
  #   column <- indivParamSummary[[i]]
  #   if (all(is.na(column))) {
  #     tempSummary <- tempSummary[, !(names(tempSummary) %in% names(indivParamSummary)[i])]
  #   }
  # }
  # indivParamSummary <- tempSummary

  # Creating the flextable with the headers
  ftParamSummary <- flextable(indivParamSummary)

  headerOffset <- 1 + bUnits + min(nbVertCat - 1, 1)
  nbHeader <- 1
  if (nbHorizCat2 == 1) {
    ftParamSummary <- ftParamSummary %>%
      add_header_row(colwidths = c(rep(1, headerOffset), rep(nbStat, nbHorizCat)), values = c(rep("", headerOffset), displayHorizCat))
  } else {
    if (nbHorizCat == 1) {
      ftParamSummary <- ftParamSummary %>%
        add_header_row(colwidths = c(rep(1, headerOffset), rep(nbStat * nbHorizCat, nbHorizCat2)), values = c(rep("", headerOffset), displayHorizCat2))
    } else {
      ftParamSummary <- ftParamSummary %>%
        add_header_row(colwidths = c(rep(1, headerOffset), rep(nbStat, nbHorizCat * nbHorizCat2)), values = c(rep("", headerOffset), displayHorizCat)) %>%
        add_header_row(colwidths = c(rep(1, headerOffset), rep(nbStat * nbHorizCat, nbHorizCat2)), values = c(rep("", headerOffset), displayHorizCat2))
      nbHeader <- 2
    }
  }

  # Apply the display settings
  ftParamSummary <- .applySettingsTable(ftParamSummary)

  # Change the display and add the number of significative numbers
  ftParamSummary <- ftParamSummary %>%
    set_header_labels(values = labelNames) %>%
    align(part = "header", align = "center") %>%
    vline_left(border = .getThickBorder(2), part = "all") %>%
    vline(j = headerOffset:ncol(indivParamSummary), border = .getSmallBorder()) %>%
    vline(j = headerOffset + seq(from = 0, to = nbHorizCat * nbHorizCat2 * nbStat - 1, by = nbStat), border = .getThickBorder(1.5)) %>%
    vline(j = c(headerOffset, ncol(indivParamSummary)), border = .getThickBorder(2))

  # Header design
  ftParamSummary <- ftParamSummary %>%
    hline(i = 1:nbHeader, border = fp_border(color = "black", width = 2), part = "header") %>%
    hline(i = 1:nbHeader, j = 1:headerOffset, border = fp_border(width = 0), part = "header") %>%
    vline(i = nbHeader, j = ncol(indivParamSummary) - nbStat + 1, border = .getThickBorder(1.5), part = "header")
  if (nbHeader == 2) {
    ftParamSummary <- ftParamSummary %>%
      vline(i = 1, j = headerOffset + (ncol(indivParamSummary) - headerOffset) / nbHorizCat2 + 1, border = .getThickBorder(1.5), part = "header")
  }

  # Add horizontal lines in case of vertCat
  if (nbVertCat > 1) {
    ftParamSummary <- ftParamSummary %>%
      bold(j = 1, bold = TRUE, part = "body") %>%
      vline(j = 1, border = .getThickBorder(1)) %>%
      vline(j = 1, border = .getThickBorder(0), part = "header") %>%
      hline(j = 1, border = .getThickBorder(0)) %>%
      hline(i = seq(from = nbParam, to = nbParam * nbVertCat, by = nbParam), border = .getThickBorder(2))
  }

  # Change the display of the variable names, units
  if (nbVertCat > 1) {
    bVert <- 1
  } else {
    bVert <- 0
  }
  ftParamSummary <- ftParamSummary %>%
    .replaceParamNamesTable(paramList = as.character(indivParamSummary[, bVert + 1]), col = bVert + 1)
  if (bUnits) {
    ftParamSummary <- ftParamSummary %>%
      .replaceUnitsTable(paramList = as.character(indivParamSummary[, bVert + 2]), col = bVert + 2)
  }

  # Add a footer if some are expressed as a median
  ftParamSummary <- .addStringFooterMedian(ftParamSummary, asMedian, stat)

  return(ftParamSummary)
}

#' Table of individual NCA metrics expressed as ratios
#'
#' \code{getNCARatioTable} returns individual NCA parameters as a Flextable
#' object in the ratio format.
#'
#' @param paramListIndiv (\emph{vector}) NCA metrics to be displayed in the
#'   table. All available options can be found in the table
#'   \href{https://pkanalix.lixoft.com/nca-parameters/}{here}.
#' @param asMedian (\emph{vector}) NCA metrics for which the calculated summary
#'   statistics is median.
#' @param stat (\emph{vector}) Summary statistics to compute. Available options
#'   are min, q1, median, q3, max, mean, sd, se, cv, cv%, geomean, geosd, geocv,
#'   geocv%, hmean and n.
#' @param nbSignDigit (\emph{int}) Number of significant digits used to describe
#'   the data.
#' @param ratioCatName (\emph{string}) Name of the categorical covariate to
#'   divide the data by.
#' @param renameRatioCat (\emph{vector}) Displayed names for each of the
#'   ratioCatName modalities.
#' @param filterRatioCat (\emph{vector}) Ratio category modalities to filter
#'   out.
#' @param horizCatName (\emph{string}) Categorical covariate to split the data
#'   by horizontally.
#' @param renameHorizCat (\emph{vector}) Displayed names for each of the
#'   horizCatName modalities.
#' @param filterHorizCat (\emph{vector}) horizCatName modalities to filter out
#'   of the table.
#' @param vertCatName (\emph{string}) Categorical covariate to split the data by
#'   vertically.
#' @param renameVertCat (\emph{vector}) Displayed names for each vertCatName
#'   modality.
#' @param filterVertCat (\emph{vector}) vertCatName modalities to filter out of
#'   the table.
#' @param vertCat2Name (\emph{string}) Categorical covariate to split the data
#'   by vertically.
#' @param renameVertCat2 (\emph{vector}) Displayed names for each vertCat2Name
#'   modality.
#' @param filterVertCat2 (\emph{vector}) vertCat2Name modalities to filter out
#'   of the table.
#' @param ratioGain (\emph{double}) A factor to multiply ratios with (e.g. 100,
#'   if you want the ratios to be in %).
#' @param ratioRef (\emph{string}) Ratio category modality with which all other
#'   should be divided (if NULL, first modality).
#' @param filterBy (\emph{vector}) Vector containing one or more of the
#'   following strings: "Flag_AUC_%Extrap_pred", "Flag_Rsq_adjusted",
#'   "Flag_Span". Note: included acceptance options need to be set in PKanalix
#'   project in order for filtering to function properly.
#'
#' @return A Flextable object.
#'
#' @examples
#' \dontrun{
#' library(lixoftConnectors)
#'
#' initializeLixoftConnectors("pkanalix")
#' demoPath <- getDemoPath()
#' loadProject(paste0(demoPath, "/3.bioequivalence/project_crossover_bioequivalence.pkx"))
#' setNCASettings(
#'   partialAucTime = list(T, c(0, 24)),
#'   computedNCAParameters = c("AUCINF_obs", "Cmax", "Tmax", "Tlast")
#' )
#' runNCAEstimation()
#'
#' RATIOTABLE <- getNCARatioTable(
#'   paramListIndiv = c("AUCINF_obs", "Cmax", "AUC_0_24", "Tmax", "Tlast"),
#'   asMedian = c("Tmax", "Tlast"),
#'   stat = c("Mean", "CV%", "HMean"),
#'   nbSignDigit = 4,
#'   ratioCatName = "FORM",
#'   renameRatioCat = c("ref" = "Reference", "test" = "Test"),
#'   vertCatName = "SEQ"
#' )
#' RATIOTABLE
#' }
#'
#' @export
getNCARatioTable <- function(paramListIndiv = NULL, asMedian = NULL, stat = c("Mean", "CV%"), nbSignDigit = 3,
                             ratioCatName = NULL, renameRatioCat = NULL, filterRatioCat = NULL,
                             horizCatName = NULL, renameHorizCat = NULL, filterHorizCat = NULL,
                             vertCatName = NULL, renameVertCat = NULL, filterVertCat = NULL,
                             vertCat2Name = NULL, renameVertCat2 = NULL, filterVertCat2 = NULL,
                             ratioGain = 1, ratioRef = NULL, filterBy = NULL) {
  return(.getNCATable(
    paramListIndiv = paramListIndiv, asMedian = asMedian, stat = stat,
    nbSignDigit = nbSignDigit, displayOCC = T, horizCatName = ratioCatName,
    renameHorizCat = renameRatioCat, filterHorizCat = filterRatioCat, horizCat2Name = horizCatName,
    renameHorizCat2 = renameHorizCat, filterHorizCat2 = filterHorizCat,
    vertCatName = vertCatName, renameVertCat = renameVertCat, filterVertCat = filterVertCat,
    vertCat2Name = vertCat2Name, renameVertCat2 = renameVertCat2, filterVertCat2 = filterVertCat2,
    ratioTable = T, ratioGain = ratioGain, ratioRef = ratioRef, filterBy = filterBy
  ))
}

#' Table of individual NCA metrics
#'
#' \code{getNCAIndivTable} returns individual NCA parameters as a Flextable
#' object.
#'
#' @param paramListIndiv (\emph{vector}) NCA metrics to be displayed in the
#'   table. All available options can be found in the table
#'   \href{https://pkanalix.lixoft.com/nca-parameters/}{here}.
#' @param asMedian (\emph{vector}) NCA metrics for which the calculated summary
#'   statistics is median.
#' @param stat (\emph{vector}) Summary statistics to compute. Available options
#'   are min, q1, median, q3, max, mean, sd, se, cv, cv%, geomean, geosd, geocv,
#'   geocv%, hmean and n.
#' @param nbSignDigit (\emph{int}) Number of significant digits used to describe
#'   the data.
#' @param displayOCC (\emph{bool}) Indicate occasion as a separate column next
#'   to the ID column.
#' @param horizCatName (\emph{string}) Categorical covariate to split the data
#'   by horizontally.
#' @param renameHorizCat (\emph{vector}) Displayed names for each of the
#'   horizCatName modalities.
#' @param filterHorizCat (\emph{vector}) horizCatName modalities to filter out
#'   of the table.
#' @param horizCat2Name (\emph{string}) Categorical covariate to split the data
#'   by horizontally.
#' @param renameHorizCat2 (\emph{vector}) Displayed names for each of the
#'   horizCat2Name modalities.
#' @param filterHorizCat2 (\emph{vector}) horizCat2Name modalities to filter out
#'   of the table.
#' @param vertCatName (\emph{string}) Categorical covariate to split the data by
#'   vertically.
#' @param renameVertCat (\emph{vector}) Displayed names for each vertCatName
#'   modality.
#' @param filterVertCat (\emph{vector}) vertCatName modalities to filter out of
#'   the table.
#' @param vertCat2Name (\emph{string}) Categorical covariate to split the data
#'   by vertically.
#' @param renameVertCat2 (\emph{vector}) Displayed names for each vertCat2Name
#'   modality.
#' @param filterVertCat2 (\emph{vector}) vertCat2Name modalities to filter out
#'   of the table.
#' @param filterBy (\emph{vector}) Vector containing one or more of the
#'   following strings: "Flag_AUC_%Extrap_pred", "Flag_Rsq_adjusted",
#'   "Flag_Span". Note: included acceptance options need to be set in PKanalix
#'   project in order for filtering to function properly.
#' @param doseNormalizationUnit (\emph{string}) String containing normalization
#'   unit, if dosing is normalized.
#'
#' @return A Flextable object.
#'
#' @examples
#' \dontrun{
#' library(lixoftConnectors)
#'
#' initializeLixoftConnectors("pkanalix")
#' demoPath <- getDemoPath()
#' loadProject(paste0(demoPath, "/3.bioequivalence/project_crossover_bioequivalence.pkx"))
#' setNCASettings(
#'   partialAucTime = list(T, c(0, 24)),
#'   computedNCAParameters = c("AUCINF_obs", "Cmax", "Tmax", "Tlast")
#' )
#' runNCAEstimation()
#'
#' INDIVTABLE <- getNCAIndivTable(
#'   paramListIndiv = c("AUCINF_obs", "Cmax", "AUC_0_24", "Tmax", "Tlast"),
#'   asMedian = c("Tmax", "Tlast"),
#'   stat = c("Mean", "CV%", "HMean"),
#'   nbSignDigit = 4,
#'   horizCatName = "FORM",
#'   renameHorizCat = c("ref" = "Reference", "test" = "Test"),
#'   vertCatName = "SEQ"
#' )
#' INDIVTABLE
#' }
#'
#' @export
getNCAIndivTable <- function(paramListIndiv = NULL, asMedian = NULL, stat = c("Mean", "CV%"), nbSignDigit = 3, displayOCC = T,
                             horizCatName = NULL, renameHorizCat = NULL, filterHorizCat = NULL,
                             horizCat2Name = NULL, renameHorizCat2 = NULL, filterHorizCat2 = NULL,
                             vertCatName = NULL, renameVertCat = NULL, filterVertCat = NULL,
                             vertCat2Name = NULL, renameVertCat2 = NULL, filterVertCat2 = NULL,
                             filterBy = NULL, doseNormalizationUnit = NULL) {
  return(.getNCATable(paramListIndiv, asMedian, stat, nbSignDigit, displayOCC,
    horizCatName, renameHorizCat, filterHorizCat,
    horizCat2Name, renameHorizCat2, filterHorizCat2,
    vertCatName, renameVertCat, filterVertCat,
    vertCat2Name, renameVertCat2, filterVertCat2,
    filterBy = filterBy, doseNormalizationUnit = doseNormalizationUnit
  ))
}

.getNCATable <- function(paramListIndiv = NULL, asMedian = NULL, stat = c("Mean", "CV%"), nbSignDigit = 3, displayOCC = T,
                         horizCatName = NULL, renameHorizCat = NULL, filterHorizCat = NULL,
                         horizCat2Name = NULL, renameHorizCat2 = NULL, filterHorizCat2 = NULL,
                         vertCatName = NULL, renameVertCat = NULL, filterVertCat = NULL,
                         vertCat2Name = NULL, renameVertCat2 = NULL, filterVertCat2 = NULL,
                         ratioTable = F, ratioGain = 1, ratioRef = NULL, filterBy = NULL, doseNormalizationUnit = NULL) {

  #-----------------------------------------------------------------------------
  # Check the parameters
  #-----------------------------------------------------------------------------
  if (is.null(paramListIndiv)) {
    param <- row.names(.lixoftCall("getNCAIndividualParameters")$information)
  } else {
    param <- .checkParamNames(paramListIndiv)
    if (is.logical(param)) {
      return(invisible(FALSE))
    }
    param <- sort(param)
  }
  indivParam <- .lixoftCall("getNCAIndividualParameters", list(param))$parameters

  # Filter indivParam by acceptance criteria
  if (length(filterBy) > 0) {
    for (criteria in filterBy) {
      if (!(criteria %in% names(indivParam))) {
        message(paste0("WARNING: data won't be filtered by '", criteria, "', since the parameter of that name doesn't exist."))
        next
      }
      indivParam <- indivParam[indivParam[criteria] == 1, ]
    }
  }

  # Add transformed covariate if needed
  indivParam <- .addTransformedCov(indivParam)
  nbParam <- length(param)

  #-----------------------------------------------------------------------------
  # Check that the parameters is asMedian are part of the computed datas
  #-----------------------------------------------------------------------------
  if (is.logical(.checkAsMedianParam(asMedian, param))) {
    return(invisible(FALSE))
  }

  #-----------------------------------------------------------------------
  # Check the stats
  #-----------------------------------------------------------------------
  if (is.logical(.checkStatNames(stat))) {
    return(invisible(FALSE))
  }
  nbStat <- length(stat)

  #-----------------------------------------------------------------------------
  # Check nbSignDigit, displayOCC, ratioTable, ratioGain, ratioRef
  #-----------------------------------------------------------------------------
  if (!(is.numeric(nbSignDigit) & (nbSignDigit %% 1 == 0))) {
    message("ERROR: 'nbSignDigit' must be an integer.")
    return(invisible(FALSE))
  }
  if (!is.logical(displayOCC)) {
    message("ERROR: 'displayOCC' must be an boolean.")
    return(invisible(FALSE))
  }
  if (!is.numeric(ratioGain)) {
    message("ERROR: 'ratioGain' must be a double.")
    return(invisible(FALSE))
  }
  if (!is.logical(ratioTable)) {
    message("ERROR: 'ratioTable' must be an boolean.")
    return(invisible(FALSE))
  }

  #-----------------------------------------------------------------------------
  # Get the check of the categories
  #-----------------------------------------------------------------------------
  if (is.logical(.checkCOVName(horizCatName))) {
    return(invisible(FALSE))
  }
  nbHorizCatOriginal <- .getNbModalities(horizCatName)
  if (!.checkCovAvailability(horizCatName, renameHorizCat, filterHorizCat, "horizCatName", "renameHorizCat", "filterHorizCat")) {
    return(invisible(FALSE))
  }

  if (is.logical(.checkCOVName(horizCat2Name))) {
    return(invisible(FALSE))
  }
  nbHorizCat2Original <- .getNbModalities(horizCat2Name)
  if (!.checkCovAvailability(horizCat2Name, renameHorizCat2, filterHorizCat2, "horizCat2Name", "renameHorizCat2", "filterHorizCat2")) {
    return(invisible(FALSE))
  }

  if (is.logical(.checkCOVName(vertCatName))) {
    return(invisible(FALSE))
  }
  nbVertCatOriginal <- .getNbModalities(vertCatName)
  if (!.checkCovAvailability(vertCatName, renameVertCat, filterVertCat, "vertCatName", "renameVertCat", "filterVertCat")) {
    return(invisible(FALSE))
  }

  if (is.logical(.checkCOVName(vertCat2Name))) {
    return(invisible(FALSE))
  }
  nbVertCat2Original <- .getNbModalities(vertCat2Name)
  if (!.checkCovAvailability(vertCat2Name, renameVertCat2, filterVertCat2, "vertCat2Name", "renameVertCat2", "filterVertCat2")) {
    return(invisible(FALSE))
  }

  #-----------------------------------------------------------------------
  # Check the possibility of generating ratio
  # Several modalities are needed
  #-----------------------------------------------------------------------
  if (ratioTable) {
    if (is.null(horizCatName)) {
      message(paste0("ERROR: Generating a ratio table is not possible, as 'ratioCatName' is NULL."))
      return(invisible(FALSE))
    } else {
      if (nbHorizCatOriginal == 1) {
        message(paste0("ERROR: Generating a ratio table is not possible, as '", horizCatName, "' has only one modality."))
        return(invisible(FALSE))
      } else {
        if (is.logical(.checkCOVName(horizCatName, ratioRef))) {
          return(invisible(FALSE))
        }
      }
    }
  }

  # ===========================================================================================
  # Update the categories modalities to take renaming into account
  # ===========================================================================================
  nbOcc <- sum(.lixoftCall("getData")$headerTypes == "occ")

  hCat <- .updateModalities(horizCatName, renameHorizCat, filterHorizCat)
  if (is.logical(hCat)) {
    return(invisible(FALSE))
  }
  modHorizCat <- hCat[[1]]
  displayHorizCat <- hCat[[2]]
  nbHorizCat <- length(modHorizCat)

  if (ratioTable & (nbHorizCat == 1)) {
    message(paste0("ERROR: Generating a ratio table is not possible, as '", horizCatName, "' has only one modality left after filtering."))
    return(invisible(FALSE))
  }
  if (ratioTable) {
    if (is.null(ratioRef)) {
      ratioRef <- modHorizCat[1]
    }
    if (ratioRef %in% names(renameHorizCat)) {
      displayRatioRef <- renameHorizCat[[ratioRef]]
    } else {
      displayRatioRef <- ratioRef
    }
    iRatio <- which(modHorizCat == ratioRef)
    modHorizCat <- modHorizCat[-iRatio]
    displayHorizCat <- paste0("Ratio (", displayHorizCat, " : ", displayRatioRef, ")")
    nbHorizCat <- nbHorizCat - 1
  }

  hCat2 <- .updateModalities(horizCat2Name, renameHorizCat2, filterHorizCat2)
  if (is.logical(hCat2)) {
    return(invisible(FALSE))
  }
  modHorizCat2 <- hCat2[[1]]
  displayHorizCat2 <- hCat2[[2]]
  nbHorizCat2 <- length(modHorizCat2)

  vCat <- .updateModalities(vertCatName, renameVertCat, filterVertCat)
  if (is.logical(vCat)) {
    return(invisible(FALSE))
  }
  modVertCat <- vCat[[1]]
  displayVertCat <- vCat[[2]]
  nbVertCat <- length(modVertCat)

  vCat2 <- .updateModalities(vertCat2Name, renameVertCat2, filterVertCat2)
  if (is.logical(vCat2)) {
    return(invisible(FALSE))
  }
  modVertCat2 <- vCat2[[1]]
  displayVertCat2 <- vCat2[[2]]
  nbVertCat2 <- length(modVertCat2)

  #-----------------------------------------------------------------------------
  # Compute the stats by modalities of vertCatName
  #-----------------------------------------------------------------------------
  df_Summary <- NULL
  dataIndiv <- NULL
  df_split <- df_splitVert2 <- NULL
  for (indexVert2 in 1:nbVertCat2) {
    # get the index (ID#OCC) of the needed
    indexFilter_Vert2 <- 1:nrow(indivParam)
    if (nbVertCat2Original > 1) {
      colVal <- indivParam[[vertCat2Name]][indexFilter_Vert2]
      indexFilter_Vert2 <- indexFilter_Vert2[which(colVal == modVertCat2[indexVert2])]
    }
    nbdf_splitVert2 <- 0
    for (indexVert in 1:nbVertCat) {
      # get the index (ID#OCC) of the needed
      indexFilter <- indexFilter_Vert2
      if (nbVertCatOriginal > 1) {
        colVal <- indivParam[[vertCatName]][indexFilter]
        indexFilter <- indexFilter[which(colVal == modVertCat[indexVert])]
      }
      df_Summary <- NULL
      indexFilterVertRef <- indexFilter
      # Get all the IDs of the cohort
      IDs <- as.character(unique(indivParam$ID[indexFilterVertRef]))
      # Reorder the IDs in case of integers
      if (sum(suppressWarnings(is.na(as.double(IDs)))) == 0) {
        IDs <- sort(as.double(IDs))
      }
      if ((nbHorizCatOriginal == 0) & (nbHorizCat2Original == 0)) {
        df_Summary <- indivParam[indexFilter, (1 + nbOcc) + (1:nbParam)]

        # Reorder in case of
        if (sum(suppressWarnings(is.na(as.double(as.character(indivParam[indexFilter, 1]))))) == 0) {
          indexFilter <- indexFilter[sort(as.double(as.character(indivParam$ID[indexFilter])), index.return = T)$ix]
        }

        if (displayOCC & (nbOcc > 0)) {
          dfName <- indivParam[indexFilter, 1:(1 + nbOcc)]
          bDisplayOcc <- 1
        } else {
          dfName <- data.frame(indivParam[indexFilter, 1])
          names(dfName) <- names(indivParam)[1]
          bDisplayOcc <- 0
        }
      } else {
        dfName <- data.frame(ID = IDs)
        names(dfName) <- names(indivParam)[1]
        bDisplayOcc <- 0
        showedOccasionWarning <- F

        for (indexHorz2 in 1:nbHorizCat2) {
          indexFilter <- indexFilterVertRef
          if (nbHorizCat2Original > 1) {
            colVal <- indivParam[[horizCat2Name]][indexFilter]
            indexFilter <- indexFilter[which(colVal == modHorizCat2[indexHorz2])]
          }
          colID <- indivParam$ID[indexFilter]

          if (!is.null(horizCatName)) {
            colCat <- indivParam[[horizCatName]][indexFilter]
          } else {
            colCat <- modHorizCat
          }

          # Generate the reference table if ratio is computed
          if (ratioTable) {
            dfHoriz_ref <- matrix(nrow = length(IDs), ncol = nbParam)
            for (indexID in 1:length(IDs)) {
              indexIDOCC <- which((colID == IDs[indexID]) & (colCat == ratioRef))
              if (length(indexIDOCC) > 1) {
                if (!showedOccasionWarning) {
                  message("Warning: several occasions correspond to the same split. Average will be used.")
                  showedOccasionWarning <- T
                }
              }
              if (length(indexIDOCC) > 0) {
                dfHoriz_ref[indexID, ] <- colMeans(as.matrix(indivParam[indexFilter[indexIDOCC], (1 + nbOcc) + (1:nbParam)]))
              }
            }
            dfHoriz_ref <- data.frame(dfHoriz_ref)
            names(dfHoriz_ref) <- param
          }

          # Generate the table with the good category
          for (indexHorz1 in 1:nbHorizCat) {
            dfHoriz <- matrix(nrow = length(IDs), ncol = nbParam)
            for (indexID in 1:length(IDs)) {
              indexIDOCC <- which((colID == IDs[indexID]) & (colCat == modHorizCat[indexHorz1]))
              if (length(indexIDOCC) > 1) {
                if (!showedOccasionWarning) {
                  message("Warning: several occasions correspond to the same split. Average will be used.")
                  showedOccasionWarning <- T
                }
              }
              if (length(indexIDOCC) > 0) {
                dfHoriz[indexID, ] <- colMeans(as.matrix(indivParam[indexFilter[indexIDOCC], (1 + nbOcc) + (1:nbParam)]))
              }
            }
            dfHoriz <- data.frame(dfHoriz)
            names(dfHoriz) <- param

            # Update the dfHoriz if it is a ratio
            if (ratioTable) {
              for (indexCol in 1:nbParam) {
                isParamAUC <- length(grep(x = param[indexCol], pattern = "AUC"))
                if (isParamAUC) {
                  dfHoriz[, indexCol] <- ratioGain * dfHoriz[, indexCol] / dfHoriz_ref[, indexCol]
                } else {
                  dfHoriz[, indexCol] <- ratioGain * dfHoriz[, indexCol] / dfHoriz_ref[, indexCol]
                }
              }
            }
            if (is.null(df_Summary)) {
              df_Summary <- dfHoriz
            } else {
              df_Summary <- cbind(df_Summary, dfHoriz)
            }
          }
        }
      }

      df_Summary <- cbind(dfName, df_Summary)

      df_Summary[[1]] <- as.character(df_Summary[[1]])
      if (nrow(df_Summary) > 0) {
        df_Summary <- .getStatIndiv(df_Summary, stat, asMedian, nbOccDisplay = nbOcc * bDisplayOcc)

        if (is.null(dataIndiv)) {
          # if((indexVert==1)&(indexVert2==1)){
          dataIndiv <- df_Summary
        } else {
          dataIndiv <- rbind(dataIndiv, df_Summary)
        }
        if (nrow(df_Summary) > 0) {
          df_split <- c(df_split, displayVertCat[indexVert], rep("", nrow(df_Summary) - 1))
        }
        nbdf_splitVert2 <- nbdf_splitVert2 + nrow(df_Summary)
      }
    }
    df_splitVert2 <- c(df_splitVert2, displayVertCat2[indexVert2], rep("", nbdf_splitVert2 - 1))
  }

  # If there is a vertical split, add the categories on the left
  if (nbVertCat > 1) {
    df_split <- data.frame(df_split)
    names(df_split) <- vertCatName
    dataIndiv <- cbind(df_split, dataIndiv)
    bVertCat <- 1
  } else {
    bVertCat <- 0
  }

  # If there is a vertical split, add the categories on the left
  if (nbVertCat2 > 1) {
    df_splitVert2 <- data.frame(df_splitVert2)
    names(df_splitVert2) <- vertCat2Name
    dataIndiv <- cbind(df_splitVert2, dataIndiv)
    bVertCat2 <- 1
  } else {
    bVertCat2 <- 0
  }

  # Definition of the offset for the display of the vertCatName, vertCat2Name, ID, Occs, ..
  offset <- 1 + bDisplayOcc * nbOcc + bVertCat + bVertCat2

  #-----------------------------------------------------------------------------
  # Clean the data
  #-----------------------------------------------------------------------------
  for (indexCol in (1 + offset):ncol(dataIndiv)) {
    dataIndiv[, indexCol] <- suppressWarnings(as.double(dataIndiv[, indexCol]))
  }

  # Update the labels to "remove" the fake names
  initDataheader <- names(dataIndiv)
  labelNames <- list()
  for (index in 1:length(initDataheader)) {
    labelNames[index] <- initDataheader[index]
  }
  names(labelNames) <- paste0(initDataheader, "_", 1:length(initDataheader))
  # labelNames$ID = getData()$header[which(getData()$headerTypes=="id")]

  names(dataIndiv) <- names(labelNames)
  # Get the units
  unitsLine <- as.character(.lixoftCall("getNCAIndividualParameters", list(param))$information$units)
  if (length(unitsLine) == 0) {
    message("Warning: Units were not used in the PKanalix project.")
    bUnits <- 0
  } else {
    bUnits <- 1
    if (!is.null(doseNormalizationUnit)) {
      for (i in 1:length(param)) {
        unitsLine[i] <- paste0(unitsLine[i], .addNormalization(param[i], doseNormalizationUnit))
      }
    }
  }
  if (ratioTable) {
    bUnits <- 0
  }
  # Do not get the round of the ID (as it could be an integer)
  dataIndiv[, (1 + offset):ncol(dataIndiv)] <- .getRound(dataIndiv[, (1 + offset):ncol(dataIndiv)], nbSignDigit)
  ftParam <- flextable(dataIndiv) %>%
    set_header_labels(values = labelNames) %>%
    align(part = "header", align = "center")

  rep_Offset <- rep(1, offset)
  rep_OffsetString <- rep("", offset)

  nbHeader <- 0
  if ((nbHorizCat > 1) & (nbHorizCat2 == 1)) {
    if (ratioTable) {
      ftParam <- ftParam %>%
        add_header_row(colwidths = c(rep_Offset, rep(nbParam, nbHorizCat)), values = c(rep_OffsetString, displayHorizCat[-iRatio]), top = T)
    } else {
      ftParam <- ftParam %>%
        add_header_row(colwidths = c(rep_Offset, rep(nbParam, nbHorizCat)), values = c(rep_OffsetString, displayHorizCat), top = T)
    }
    nbHeader <- 1
  } else if ((nbHorizCat2 > 1) & (nbHorizCat == 1)) {
    ftParam <- ftParam %>%
      add_header_row(colwidths = c(rep_Offset, rep(nbParam, nbHorizCat2)), values = c(rep_OffsetString, displayHorizCat2), top = T)
    nbHeader <- 1
  } else if ((nbHorizCat2 > 1) & (nbHorizCat > 1)) {
    if (ratioTable) {
      ftParam <- ftParam %>%
        add_header_row(colwidths = c(rep_Offset, rep(nbParam, nbHorizCat * nbHorizCat2)), values = c(rep_OffsetString, rep(displayHorizCat[-iRatio], nbHorizCat2)), top = T) %>%
        add_header_row(colwidths = c(rep_Offset, rep(nbParam * nbHorizCat, nbHorizCat2)), values = c(rep_OffsetString, displayHorizCat2), top = T)
    } else {
      ftParam <- ftParam %>%
        add_header_row(colwidths = c(rep_Offset, rep(nbParam, nbHorizCat * nbHorizCat2)), values = c(rep_OffsetString, rep(displayHorizCat, nbHorizCat2)), top = T) %>%
        add_header_row(colwidths = c(rep_Offset, rep(nbParam * nbHorizCat, nbHorizCat2)), values = c(rep_OffsetString, displayHorizCat2), top = T)
    }
    nbHeader <- 2
  }

  if (bUnits) {
    ftParam <- ftParam %>%
      add_header_row(values = c(rep_OffsetString, rep(unitsLine, nbHorizCat * nbHorizCat2)), top = F)
  }
  ftParam <- ftParam %>%
    theme_vanilla() %>%
    align(align = "center") %>%
    align(part = "header", align = "center") %>%
    font(part = "all", fontname = "Calibri")

  # Body design
  ftParam <- ftParam %>%
    vline_left(border = .getThickBorder(2), part = "all") %>%
    vline(j = offset:ncol(dataIndiv), border = .getSmallBorder()) %>%
    vline(j = offset + seq(from = 0, to = nbHorizCat2 * nbHorizCat * nbParam - 1, by = nbParam), border = .getThickBorder(1.5)) %>%
    vline(j = c(offset, ncol(dataIndiv)), border = .getThickBorder(2))

  # Header design
  if (nbHeader + bUnits > 0) {
    if (ratioTable) {
      ftParam <- ftParam %>%
        hline(i = nbHeader, border = fp_border(width = 0), part = "header")
    } else {
      ftParam <- ftParam %>%
        hline(i = nbHeader + 1, border = fp_border(width = 0), part = "header")
    }
  }
  if (nbHeader > 0) {
    ftParam <- ftParam %>%
      hline(i = 1:nbHeader, border = fp_border(color = "black", width = 2), part = "header") %>%
      hline(i = 1:nbHeader, j = 1:offset, border = fp_border(width = 0), part = "header") %>%
      vline(i = 1:nbHeader, j = ncol(dataIndiv) - nbParam + 1, border = .getThickBorder(1.5), part = "header")
  }

  if (nbVertCat > 1) {
    bVertCat <- 1
  } else {
    bVertCat <- 0
  }
  if (nbVertCat2 > 1) {
    bVertCat2 <- 1
  } else {
    bVertCat2 <- 0
  }

  # Display of the vertCat
  if (bVertCat + bVertCat2 > 0) {
    ftParam <- ftParam %>%
      bold(j = 1:(bVertCat + bVertCat2), bold = TRUE, part = "body")
    if (bVertCat + bVertCat2 > 1) {
      ftParam <- ftParam %>%
        vline(j = 1, border = .getThickBorder(2)) %>%
        vline(j = 2, border = .getThickBorder(1)) %>%
        vline(j = 1, border = .getThickBorder(0), part = "header") %>%
        vline(j = 2, border = .getThickBorder(0), part = "header") %>%
        hline(j = 1, border = .getThickBorder(0)) %>%
        hline(j = 2, border = .getThickBorder(0))
    } else {
      ftParam <- ftParam %>%
        vline(j = 1, border = .getThickBorder(1)) %>%
        vline(j = 1, border = .getThickBorder(0), part = "header") %>%
        hline(j = 1, border = .getThickBorder(0))
    }
  }
  # Put the stat in bold
  lineBold <- which(is.element(dataIndiv[[offset]], .renameStat(stat)))
  if (length(lineBold) > 0) {
    ftParam <- ftParam %>%
      hline(i = lineBold[seq(from = 1, to = length(lineBold), by = nbStat)] - 1, border = .getThickBorder()) %>% # +2 to add Name and units
      hline(i = lineBold[seq(from = nbStat, to = length(lineBold), by = nbStat)], border = .getThickBorder()) %>% # +2 to add Name and units
      bold(i = lineBold, bold = TRUE, part = "body")

    if (bVertCat + bVertCat2 > 0) {
      ftParam <- ftParam %>%
        hline(i = lineBold[seq(from = 1, to = length(lineBold), by = nbStat)] - 1, j = 1:(bVertCat + bVertCat2), border = .getThickBorder(0))
    }
    if (bVertCat + bVertCat2 > 1) {
      ftParam <- ftParam %>%
        hline(i = lineBold[seq(from = nbStat, to = length(lineBold), by = nbStat)], j = 1, border = .getThickBorder(0))
    }
  }

  if (bVertCat + bVertCat2 > 0) {
    ind_col1 <- which(dataIndiv[, 1] != "")
    ftParam <- ftParam %>%
      hline(i = c(ind_col1[2:length(ind_col1)] - 1, nrow(dataIndiv)), border = .getThickBorder())
  }

  if (bVertCat + bVertCat2 == 2) {
    ind_col2 <- which(dataIndiv[, 2] != "")
    ftParam <- ftParam %>%
      hline(i = c(ind_col2[2:length(ind_col2)] - 1), j = 2:ncol(dataIndiv), border = .getThickBorder())
  }

  if (nbHorizCat > 1) {
    bHorizCat <- 1
  } else {
    bHorizCat <- 0
  }
  if (nbHorizCat2 > 1) {
    bHorizCat2 <- 1
  } else {
    bHorizCat2 <- 0
  }

  ftParam <- ftParam %>%
    .replaceParamNamesHeader(paramList = rep(as.character(param), nbHorizCat * nbHorizCat2), line = 1 + bHorizCat + bHorizCat2, col = offset)
  if (bUnits) {
    ftParam <- ftParam %>%
      .replaceUnitsHeader(paramList = rep(as.character(unitsLine), nbHorizCat * nbHorizCat2), line = 2 + bHorizCat + bHorizCat2, col = offset)
  }

  ftParam <- ftParam %>% padding(padding = 1, part = "all")

  # Add a footer if some are expressed as a median
  ftParam <- .addStringFooterMedian(ftParam, asMedian, stat)

  if (ratioTable) {
    if (nbHorizCat == 1) {
      ftParam <- add_header_row(ftParam, values = c("", displayHorizCat[-iRatio]), colwidths = c(1, length(labelNames) - 1)) %>%
        vline(i = 1, j = ncol(dataIndiv) - nbParam + 1, border = .getThickBorder(1.5), part = "header")
    }
  }

  return(ftParam)
}

#' Table of measured drug concentrations
#'
#' \code{getConcTable} returns measured concentrations as a Flextable object.
#'
#' @param stat (\emph{vector}) Summary statistics to compute. Available options
#'   are min, q1, median, q3, max, mean, sd, se, cv, cv%, geomean, geosd, geocv,
#'   geocv%, hmean and n.
#' @param nbSignDigit (\emph{int}) Number of significant digits used to describe
#'   the data.
#' @param displayOCC (\emph{bool}) Indicate occasion as a separate column next
#'   to the ID column.
#' @param vertCatName (\emph{string}) Categorical covariate to split the data by
#'   vertically.
#' @param renameVertCat (\emph{vector}) Displayed names for each vertCatName
#'   modality.
#' @param filterVertCat (\emph{vector}) vertCatName modalities to filter out of
#'   the table.
#' @param vertCat2Name (\emph{string}) Categorical covariate to split the data
#'   by vertically.
#' @param renameVertCat2 (\emph{vector}) Displayed names for each vertCat2Name
#'   modality.
#' @param filterVertCat2 (\emph{vector}) vertCat2Name modalities to filter out
#'   of the table.
#'
#' @return A Flextable object.
#'
#' @examples
#' \dontrun{
#' library(lixoftConnectors)
#'
#' initializeLixoftConnectors("pkanalix")
#' demoPath <- getDemoPath()
#' loadProject(paste0(demoPath, "/3.bioequivalence/project_crossover_bioequivalence.pkx"))
#' setNCASettings(
#'   partialAucTime = list(T, c(0, 24)),
#'   computedNCAParameters = c("AUCINF_obs", "Cmax", "Tmax", "Tlast")
#' )
#' runNCAEstimation()
#'
#' CONCTABLE <- getConcTable(
#'   vertCatName = "FORM",
#'   renameVertCat = c("ref" = "Reference", "test" = "Test")
#' )
#' CONCTABLE
#' }
#'
#' @export
getConcTable <- function(stat = c("Mean", "CV%"), nbSignDigit = 3, displayOCC = T,
                         vertCatName = NULL, renameVertCat = NULL, filterVertCat = NULL,
                         vertCat2Name = NULL, renameVertCat2 = NULL, filterVertCat2 = NULL) {


  #-----------------------------------------------------------------------
  # Check the stats
  #-----------------------------------------------------------------------
  if (is.logical(.checkStatNames(stat))) {
    return(invisible(FALSE))
  }
  nbStat <- length(stat)

  #-----------------------------------------------------------------------------
  # Check nbSignDigit, displayOCC, ratioTable, ratioGain, ratioRef
  #-----------------------------------------------------------------------------
  if (!(is.numeric(nbSignDigit) & (nbSignDigit %% 1 == 0))) {
    message("ERROR: 'nbSignDigit' must be an integer.")
    return(invisible(FALSE))
  }
  if (!is.logical(displayOCC)) {
    message("ERROR: 'displayOCC' must be an boolean.")
    return(invisible(FALSE))
  }

  #-----------------------------------------------------------------------------
  # Get the check of the categories
  #-----------------------------------------------------------------------------
  if (is.logical(.checkCOVName(vertCatName))) {
    return(invisible(FALSE))
  }
  nbVertCatOriginal <- .getNbModalities(vertCatName)
  if (!.checkCovAvailability(vertCatName, renameVertCat, filterVertCat, "vertCatName", "renameVertCat", "filterVertCat")) {
    return(invisible(FALSE))
  }

  if (is.logical(.checkCOVName(vertCat2Name))) {
    return(invisible(FALSE))
  }
  nbVertCat2Original <- .getNbModalities(vertCat2Name)
  if (!.checkCovAvailability(vertCat2Name, renameVertCat2, filterVertCat2, "vertCat2Name", "renameVertCat2", "filterVertCat2")) {
    return(invisible(FALSE))
  }

  # ===========================================================================================
  # Update the categories modalities to take renaming into account
  # ===========================================================================================
  nbOcc <- sum(.lixoftCall("getData")$headerTypes == "occ")

  vCat <- .updateModalities(vertCatName, renameVertCat, filterVertCat)
  if (is.logical(vCat)) {
    return(invisible(FALSE))
  }
  modVertCat <- vCat[[1]]
  displayVertCat <- vCat[[2]]
  nbVertCat <- length(modVertCat)

  vCat2 <- .updateModalities(vertCat2Name, renameVertCat2, filterVertCat2)
  if (is.logical(vCat2)) {
    return(invisible(FALSE))
  }
  modVertCat2 <- vCat2[[1]]
  displayVertCat2 <- vCat2[[2]]
  nbVertCat2 <- length(modVertCat2)

  #-----------------------------------------------------------------------------
  # Compute the stats by modalities of vertCatName
  #-----------------------------------------------------------------------------
  obsName <- .lixoftCall("getObservationInformation")$name
  data <- .lixoftCall("getObservationInformation")[[obsName]]
  timeVect <- sort(unique(data$time))

  df_split <- df_splitVert2 <- dataIndiv <- NULL
  indivParam <- .lixoftCall("getNCAIndividualParameters")$parameters
  indivParam <- .addTransformedCov(indivParam)
  bDisplayOcc <- displayOCC

  for (indexVert2 in 1:nbVertCat2) {
    # get the index (ID#OCC) of the needed
    indexFilter_Vert2 <- 1:nrow(indivParam)
    if (nbVertCat2Original > 1) {
      colVal <- indivParam[[vertCat2Name]][indexFilter_Vert2]
      indexFilter_Vert2 <- indexFilter_Vert2[which(colVal == modVertCat2[indexVert2])]
    }

    for (indexVert in 1:nbVertCat) {
      # get the index (ID#OCC) of the needed
      indexFilter <- indexFilter_Vert2
      if (nbVertCatOriginal > 1) {
        colVal <- indivParam[[vertCatName]][indexFilter]
        indexFilter <- indexFilter[which(colVal == modVertCat[indexVert])]
      }
      df <- matrix(nrow = length(indexFilter), ncol = 1 + nbOcc + length(timeVect))
      for (indexID in 1:length(indexFilter)) {
        idOCC_ref <- as.character(indivParam[indexFilter[indexID], 1:(1 + nbOcc)])
        df[indexID, 1:(1 + nbOcc)] <- idOCC_ref
        validData <- rep(T, nrow(data))
        for (indexCol in 1:(1 + nbOcc)) {
          validData <- validData & (data[, indexCol] == idOCC_ref[indexCol])
        }
        indexValid <- which(validData)
        for (indexTime in 1:length(timeVect)) {
          iData <- which(data$time[indexValid] == timeVect[indexTime])
          if (length(iData) == 0) {
            df[indexID, indexTime + 1 + nbOcc] <- NA
          } else {
            if (!is.null(data$censored) && data$censored[indexValid[iData]]) {
              df[indexID, indexTime + 1 + nbOcc] <- "LOQ"
            } else {
              df[indexID, indexTime + 1 + nbOcc] <- data[[obsName]][indexValid[iData[1]]]
            }
          }
        }
      }
      df <- data.frame(df)
      names(df) <- c(names(indivParam)[1:(1 + nbOcc)], timeVect)

      for (indexCol in 1:ncol(df)) {
        df[, indexCol] <- as.character(df[, indexCol])
      }
      if (!bDisplayOcc) {
        df <- df[, -(2:(1 + nbOcc))]
      }
      if (length(stat) > 0) {
        df <- .getStatIndiv(df, stat, asMedian = NULL, nbOccDisplay = nbOcc * bDisplayOcc)
      }
      if ((indexVert == 1) & (indexVert2 == 1)) {
        dataIndiv <- df
      } else {
        dataIndiv <- rbind(dataIndiv, df)
      }
      df_split <- c(df_split, displayVertCat[indexVert], rep("", nrow(df) - 1))
    }
    df_splitVert2 <- c(df_splitVert2, displayVertCat2[indexVert2], rep("", length(indexFilter_Vert2) - 1 + 2 * length(stat)))
  }

  # If there is a vertical split, add the categories on the left
  if (nbVertCat > 1) {
    df_split <- data.frame(df_split)
    names(df_split) <- vertCatName
    dataIndiv <- cbind(df_split, dataIndiv)
    bVertCat <- 1
  } else {
    bVertCat <- 0
  }

  # If there is a vertical split, add the categories on the left
  if (nbVertCat2 > 1) {
    df_splitVert2 <- data.frame(df_splitVert2)
    names(df_splitVert2) <- vertCat2Name
    dataIndiv <- cbind(df_splitVert2, dataIndiv)
    bVertCat2 <- 1
  } else {
    bVertCat2 <- 0
  }

  # Definition of the offset for the display of the vertCatName, vertCat2Name, ID, Occs, ..
  offset <- 1 + bDisplayOcc * nbOcc + bVertCat + bVertCat2

  #-----------------------------------------------------------------------------
  # Clean the data
  #-----------------------------------------------------------------------------

  # for(indexCol in (1+offset):ncol(dataIndiv)){
  #   dataIndiv[,indexCol]<- suppressWarnings(as.double(dataIndiv[,indexCol]))
  # }

  # Update the labels to "remove" the fake names
  initDataheader <- names(dataIndiv)
  labelNames <- list()
  for (index in 1:length(initDataheader)) {
    labelNames[index] <- initDataheader[index]
  }
  names(labelNames) <- initDataheader
  # labelNames$ID = getData()$header[which(getData()$headerTypes=="id")]


  # Do not get the round of the ID (as it could be an integer)
  dataIndiv[, (1 + offset):ncol(dataIndiv)] <- .getRound(dataIndiv[, (1 + offset):ncol(dataIndiv)], nbSignDigit)
  ftParam <- flextable(dataIndiv) %>%
    set_header_labels(values = labelNames) %>%
    align(part = "header", align = "center")

  ftParam <- ftParam %>%
    theme_vanilla() %>%
    align(align = "center") %>%
    align(part = "header", align = "center") %>%
    font(part = "all", fontname = "Calibri") %>%
    hline(i = 1, border = .getThickBorder(), part = "header") %>%
    vline(j = ncol(dataIndiv), border = .getThickBorder()) %>%
    # hline(j=1:ncol(dataIndiv), border = .getThickBorder(), part = "header") %>%
    vline(j = offset, border = .getThickBorder()) %>%
    vline(j = offset + 1:length(timeVect), border = .getSmallBorder())

  lineBold <- which(is.element(dataIndiv[[offset]], stat))
  if (length(lineBold) > 0) {
    ftParam <- ftParam %>%
      hline(i = lineBold[seq(from = 1, to = length(lineBold), by = nbStat)] - 1, border = .getThickBorder()) %>% # +2 to add Name and units
      hline(i = lineBold[seq(from = 2, to = length(lineBold), by = nbStat)], border = .getThickBorder()) %>% # +2 to add Name and units
      bold(i = lineBold, bold = TRUE, part = "body")
  }

  if (nbVertCat > 1) {
    ftParam <- ftParam %>% bold(j = 1, bold = TRUE, part = "body")
  }

  ftParam <- ftParam %>%
    padding(padding = 1, part = "all") %>%
    vline_left(border = .getThickBorder(2), part = "all") %>%
    vline(j = ncol(dataIndiv), border = .getThickBorder()) %>%
    vline(j = 1, border = .getThickBorder())

  return(ftParam)
}

#' Bioequivalence summary table
#'
#' \code{getBETable} returns a table containing bioequivalence results as a
#' Flextable object.
#'
#' @param renameFormulations (\emph{vector}) Displayed names of different
#'   formulations.
#' @param doseNormalizationUnit (\emph{string}) A string containing dose
#'   normalization unit, if dosing is normalized.
#' @param test (\emph{string}) In case of multiple test formulations, test
#'   formulation for which bioequivalence table will be returned.
#'
#' @return A Flextable object.
#'
#' @examples
#' \dontrun{
#' library(lixoftConnectors)
#'
#' initializeLixoftConnectors("pkanalix")
#' demoPath <- getDemoPath()
#' loadProject(paste0(demoPath, "/3.bioequivalence/project_crossover_bioequivalence.pkx"))
#' setNCASettings(
#'   partialAucTime = list(T, c(0, 24)),
#'   computedNCAParameters = c("AUCINF_obs", "Cmax", "Tmax", "Tlast")
#' )
#' runNCAEstimation()
#' runBioequivalenceEstimation()
#'
#' BETABLE <- getBETable(renameFormulations = c("1" = "Reference", "2" = "Test"))
#' BETABLE
#' }
#'
#' @export
getBETable <- function(renameFormulations = NULL, doseNormalizationUnit = NULL, test = NULL) {
  nameOcc <- .lixoftCall("getData")$header[which(.lixoftCall("getData")$headerTypes == "occ")]

  .lixoftCall("runNCAEstimation")
  .lixoftCall("runBioequivalenceEstimation")
  CI <- .lixoftCall("getBioequivalenceResults")$confidenceIntervals
  if (!is.null(test) && test %in% names(CI)) {
    CI <- CI[[test]]
  } else {
    if (length(CI) > 1) {
      message("NOTE: multiple test formulations detected, first one will be used. You can specify the test formulation with the test argument.")
    }
    CI <- CI[[1]]
  }

  unitsEnabled <- .lixoftCall("getDataSettings")$enableunits

  # columns kept Parameter, adjR2 and N for both and the info
  CI <- CI[, c(1, 2:3, 4:5, 9:11)]
  if (unitsEnabled) {
    CI <- cbind(
      CI,
      data.frame(
        Units = as.character(
          .lixoftCall(
            "getNCAParameterStatistics",
            list(as.character(CI$Parameter))
          )$information$units
        )
      )
    )
  }
  if (unitsEnabled) {
    CI <- CI[, c(1, 9, 2:8)]
    if (!is.null(doseNormalizationUnit)) {
      for (i in 1:length(CI$Parameter)) {
        CI$Units[i] <- paste0(CI$Units[i], .addNormalization(CI$Parameter[i], doseNormalizationUnit))
      }
    }
  }

  # modCat <- as.character(unlist(unique(.lixoftCall("getNCAIndividualParameters")$parameters[nameOcc])))
  form <- .lixoftCall("getBioequivalenceSettings")$linearmodelfactors$formulation
  modCat <- as.character(unlist(unique(.lixoftCall("getNCAIndividualParameters")$parameters[form])))
  dispalayCat <- c()
  dispalayCat[1] <- names(.lixoftCall("getBioequivalenceResults")$confidenceIntervals)[1]
  dispalayCat[2] <- .lixoftCall("getBioequivalenceSettings")$linearmodelfactors$reference
  if (!is.null(renameFormulations)) {
    for (i in 1:length(dispalayCat)) {
      cat <- dispalayCat[i]
      if (!is.na(renameFormulations[cat])) {
        dispalayCat[i] <- renameFormulations[cat]
      }
    }
  }

  if (length(grep(modCat[1], names(CI)[3])) == 0) { # The modalities are not the good ones
    modCat <- modCat[c(2, 1)]
  }

  # Update the labels
  labelNames <- list()
  for (index in 1:length(names(CI))) {
    nameCol <- names(CI)[index]
    labelNames[index] <- strsplit(nameCol, split = "_", fixed = T)[[1]][1]
    if (labelNames[index] == "Ratio") labelNames[index] <- "Ratio (%)"
    if (labelNames[index] == "CILower") labelNames[index] <- "CIlower (%)"
    if (labelNames[index] == "CIUpper") labelNames[index] <- "CIupper (%)"
    if (labelNames[index] == "AdjustedMean") labelNames[index] <- "LMS"
  }
  names(labelNames) <- names(CI)

  # Creating the flextable with the headers
  if (unitsEnabled) {
    ftBE <- flextable(CI) %>%
      add_header_row(colwidths = c(c(1, 1), c(2, 2, 3)), values = c("", "", as.character(dispalayCat), "Bioequivalence"))
  } else {
    ftBE <- flextable(CI) %>%
      add_header_row(colwidths = c(c(1), c(2, 2, 3)), values = c("", as.character(dispalayCat), "Bioequivalence"))
  }


  # Apply the settings
  ftBE <- .applySettingsTable(ftBE)

  if (unitsEnabled) {
    jDoubles <- c(3, 5, 7, 8, 9)
    jLines <- c(2, 4, 6)
  } else {
    jDoubles <- c(2, 4, 6, 7, 8)
    jLines <- c(1, 3, 5)
  }
  ftBE <- ftBE %>%
    set_header_labels(values = labelNames) %>%
    align(part = "header", align = "center") %>%
    colformat_double(i = 1:nrow(CI), j = jDoubles, digits = 2) %>%
    vline(j = jLines, border = .getSmallBorder()) %>%
    .replaceParamNamesTable(paramList = as.character(CI[, 1]))
  if (unitsEnabled) {
    ftBE <- .replaceUnitsTable(ftBE, paramList = as.character(CI[, 2]))
  }

  return(ftBE)
}


# =========================================================================================================
# Add the String as a footer is needed due to the median
# =========================================================================================================
.addStringFooterMedian <- function(ft, asMedian, stat) {
  changedToMedian <- intersect(c("mean", "geomean"), tolower(stat))
  if (!is.null(asMedian) && length(changedToMedian) > 0) {
    stringFooter <- "as_paragraph("
    for (i in 1:(length(asMedian))) {
      stringFooter <- paste0(stringFooter, .renameParamNames(asMedian[i]), ", ")
      if (length(asMedian) > 1 && i == length(asMedian)-1) {
        stringFooter <- paste0(stringFooter, "as_chunk(' and '),")
      }
      if (length(asMedian) > 1 && i < length(asMedian)-1) {
        stringFooter <- paste0(stringFooter, "as_chunk(', '),")
      }
    }
    if (length(asMedian) > 0) {
      stringFooter <- paste0(stringFooter, "as_chunk(' '),")
    }
    for (i in 1:length(changedToMedian)) {
      stringFooter <- paste0(stringFooter, "as_chunk('", .renameStat(changedToMedian[i]), "'), ")
      if (length(changedToMedian) > 1 && i == length(changedToMedian)-1) {
        stringFooter <- paste0(stringFooter, "as_chunk(' and '),")
      }
      if (length(changedToMedian) > 1 && i < length(changedToMedian)-1) {
        stringFooter <- paste0(stringFooter, "as_chunk(', '),")
      }
    }
    stringFooter <- paste0(stringFooter, "as_chunk(' expressed as median.')")
    stringFooter <- paste0(stringFooter, ")")
    ft <- ft %>%
      add_footer_lines("") %>%
      compose(i = 1, j = 1, value = eval(parse(text = stringFooter)), part = "footer")
  }
  return(ft)
}
