###############################################################################
# Get value with good significative number
###############################################################################
# Get the significative number of a vector
sigfig <- function(x, digits, reporting = T){

  xout <- x
  iNaN <- which(is.nan(xout))
  if(length(iNaN)>0){
    xout[iNaN]<- NA
  }
  xNum <- suppressWarnings(as.double(as.character(xout)))
  iIsNum <- which((!is.na(xNum))&(!is.infinite((xNum))))
  if(length(iIsNum)>0){
    if(reporting){
      xout[iIsNum] <- gsub("\\.$", "", formatC(signif(xNum[iIsNum], digits=digits), digits=digits, format="fg", flag="#"))
    }else{
      xout[iIsNum] <- signif(xNum[iIsNum], digits)
    }
  }
  return(xout)
}

# Get the significative number of a vector or a data.frame
.getRound <- function(x, digits = 2, reporting=T){

  df <- x
  if(is.data.frame(x)){
    for(index in 1:length(df)){
      df[,index] <- sigfig(x[,index], digits, reporting)
    }
  }else{
    df <- sigfig(x, digits)
  }
  return(df)
}



.getFileExt <- function(filename) {
  ext <- utils::tail(strsplit(filename, "\\.")[[1]], n=1)
  if (ext == filename) ext <- NULL
  return(ext)
}

.findFileName <- function(filename, name = NULL, ext = NULL) {
  currext <- .getFileExt(filename)
  if (is.null(ext)) {
    ext <- currext
  }
  file <- tools::file_path_sans_ext(filename)
  if (!is.null(name)) {
    outputFilename <- paste0(file, "_", name, ".", ext)
  } else {
    outputFilename <- paste0(file, ".", ext)
  }
  validName <- ! file.exists(outputFilename)
  i <- 1
  newOutputFilename <- outputFilename
  while (!validName) {
    newOutputFilename <- paste0(
      tools::file_path_sans_ext(outputFilename),
      "_", i,
      ".", ext
    )
    validName <- ! file.exists(newOutputFilename)
    i <- i + 1
  }
  return(newOutputFilename)
}

###############################################################
# Get the statistics of the vector data
# Replace it by the median if there is something in asMedian
###############################################################
.getStat<-function(data, stat, asMedian = NULL){
  statVal <- tolower(stat)
  dataOut <- data[1,]
  for(index in 1:length(dataOut)){
    vect = as.numeric(data[,index])
    iVect = which((!is.nan(vect))&(!is.na(vect)))
    if(length(iVect)>0){
      if(statVal=="min"){
        dataOut[index] = min(vect[iVect])
      }else if (statVal=="q1"){
        dataOut[index] = quantile(vect[iVect], .25)
      }else if(statVal=="median"){
        dataOut[index] = median(vect[iVect])
      }else if(statVal=="q3"){
        dataOut[index] = quantile(vect[iVect], .75)
      }else if(statVal=="max"){
        dataOut[index] = max(vect[iVect])
      }else if (statVal=="mean"){
        dataOut[index] = mean(vect[iVect])
      }else if (statVal=="sd"){
        dataOut[index] = sd(vect[iVect])
      }else if (statVal=="se"){
        dataOut[index] = sd(vect[iVect])/sqrt(length(iVect))
      }else if(statVal=="cv"){
        dataOut[index] = sd(vect[iVect])/mean(vect[iVect])
      }else if(statVal=="cv%"){
        dataOut[index] = 100*sd(vect[iVect])/mean(vect[iVect])
      }else if(statVal=="geomean"){
        dataOut[index] = exp(mean(log(vect[iVect])))
      }else if(statVal=="geosd"){
        dataOut[index] = exp(sd(log(vect[iVect])))
      }else if(statVal=="geocv"){
        dataOut[index] = exp(sd(log(vect[iVect])))/exp(mean(log(vect[iVect])))
      } else if(statVal == "geocv%") {
        dataOut[index] = 100*exp(sd(log(vect[iVect])))/exp(sd(log(vect[iVect])))
      } else if (statVal == "hmean") {
        dataOut[index] = 1 / mean(1 / vect[iVect])
      } else if (statVal == "n") {
        dataOut[index] = length(vect[iVect])
      } else{
        dataOut[index] = -1
      }
    }else{
      dataOut[index] = NA
    }
  }

  # Replace by the median if needed
  if(!is.null(asMedian)){
    for(indexMed in 1:length(asMedian)){
      paramMed <- asMedian[indexMed]
      if (statVal=="mean" || statVal == "geomean") {
        indexMed <- which(names(data) == paramMed)
        if(length(indexMed)>0){
          for(iIndexImed in 1:length(indexMed)){
            vect = as.numeric(data[,indexMed[iIndexImed]])
            iVect =  which((!is.nan(vect))&(!is.na(vect)))
            if(length(iVect)>0){
              dataOut[indexMed[iIndexImed]] <- median(vect[iVect])
            }else{
              dataOut[indexMed[iIndexImed]] <- NA
            }
          }
        }
      }else if (statVal %in% c("sd", "se", "cv", "cv%", "geosd", "geocv", "geocv%")) {
        indexMed <- which(names(data) == paramMed)
        dataOut[indexMed] <- ""
      }
    }
  }

  return(dataOut)
}

################################################################################
# Return the stat with the summary
################################################################################
.getStatSummary <- function (data, stat, asMedian = NULL, indexOffset = 0) {

  df <- NULL
  nbStat = length(stat)
  if (nbStat > 0) {
    for (indexStat in 1:length(stat)) {
      df_add <- data.frame(as.numeric(.getStat(data, stat = stat[indexStat], asMedian)))
      names(df_add) <- paste0(.renameStat(stat[indexStat]),'_',indexStat+indexOffset)
      if (is.null(df)) df <- df_add else df <- cbind(df, df_add)
    }
  }
  return(df)
}

################################################################################
# Return the stat for all the individuals for the indiv table
################################################################################
.getStatIndiv<-function(data, stat, asMedian=NULL, nbOccDisplay = 0){
  df <- data
  for(index in 1:(1+nbOccDisplay)){
    df[,index]<- as.character(df[,index])
  }
  nbStat = length(stat)
  if(nbStat>0){
    for(indexStat in 1:nbStat){
      # Add the stat as a line and rename he stat if needed
      df_add <- suppressWarnings(as.numeric(.getStat(data, stat = stat[indexStat], asMedian)))
      if(nbOccDisplay>0){
        df_add[1:(1+nbOccDisplay)]<- ""
        df_add[1+nbOccDisplay] <- .renameStat(stat[indexStat])
      }else{
        df_add[1] <- .renameStat(stat[indexStat])
      }
      df <- rbind(df, df_add)
    }
  }
  return(df)
}

################################################################################
# Update the covariate modalities to take the renameCat into account
################################################################################
.updateModalities <- function(stratCat, renameCat=NULL, filterCat=NULL){

  if(!is.null(stratCat)){
    modCat <- as.character(unique(.lixoftCall("getCovariateInformation")$covariate[[stratCat]]))
    nbCat <- length(modCat)

    # Filter the data set to take the filterCat modalities
    if(!is.null(filterCat)){
      for(indexFilt in 1 : length(filterCat)){
        if(length(intersect(modCat, filterCat[indexFilt])) == 0){
          message(paste0("ERROR: The modality '",filterCat[indexFilt] ,"', asked to be filtered, is not a modality of '",stratCat,"'."))
          return(invisible(FALSE))
        }
      }
      modCat <- modCat[modCat %in% unique(filterCat)]
      nbCat <- length(modCat)
    }

    # Rename the categories associated for the filtered data
    if(!is.null(renameCat)){
      for(indexFilt in 1 : length(names(renameCat))){
        if(length(intersect(modCat, names(renameCat)[indexFilt])) == 0){
          message(paste0("ERROR: The modality '",names(renameCat)[indexFilt] ,"', asked to be renamed, is not (anymore) a modality of '",stratCat,"'."))
          return(invisible(FALSE))
        }
      }
      if(length(intersect(modCat, names(renameCat))) < nbCat){
        message(paste0("ERROR: The renaming of categories should be a vector with ",nbCat, " values, i.e. the same as the number of modalities of ",stratCat," after filtering."))
        return(invisible(FALSE))
      }else{
        displayCat = as.vector(renameCat)
        modCat <- names(renameCat)
      }
    }else{
      displayCat = as.character(modCat)
    }
  }else{
    modCat<- "Statistics"; displayCat = modCat; nbCat = 1
  }
  return(list(modCat, displayCat))
}

################################################################################
# Update the table to add transformed covariate
################################################################################
.addTransformedCov <- function(indivParam){
  covList <- .lixoftCall("getCovariateInformation")$name
  covNotInParam <- which(!(covList %in% names(indivParam)))
  if(length(covNotInParam)>0){
    for(indexCov in 1:length(covNotInParam)){
      covName <- covList[covNotInParam[indexCov]]
      indivParam <- cbind(indivParam, data.frame(.lixoftCall("getCovariateInformation")$covariate[[covName]]))
      names(indivParam)[ncol(indivParam)] <- covName
    }
  }
  return(indivParam)
}

################################################################################
# Add the unit to a specific parameter
################################################################################
.addUnits <- function(paramStr, paramName, isExp = F) {

  unit = .lixoftCall("getNCAIndividualParameters",
                     list(paramName))$information$units
  if (length(unit) > 0) {
    paramUnit <- .renameUnits(param = as.character(unit))
    strSpaceExp = " "
    if (isExp) {
      strSpaceExp = "~~"
    }
    outString = paste0(paramStr, strSpaceExp,"(",paramUnit,")")
  } else {
    outString = paramStr
  }

  return(outString)
}

################################################################################
# Get how unit should be normalized
################################################################################
.addNormalization <- function(param, unit) {
  normUnits <- c("AURC_last_D" = unit, "Cmax_D" = paste0("*", unit), "AUClast_D" = paste0("*", unit),
                 "AUCINF_D_obs" = paste0("*", unit), "AUCINF_D_pred" = paste0("*", unit),
                 "Vz_F_obs" = paste0("/", unit), "Cl_F_obs" = paste0("/", unit),
                 "Cl_obs" = paste0("/", unit), "Cl_pred" = paste0("/", unit),
                 "Vss_obs" = paste0("/", unit), "Vz_F_pred" = paste0("/", unit),
                 "Cl_F_pred" = paste0("/", unit), "Vss_pred" = paste0("/", unit),
                 "AUC_TAU_D" = paste0("*", unit), "Vz" = paste0("/", unit),
                 "Vz_obs" = paste0("/", unit), "Vz_pred" = paste0("/", unit),
                 "Vz_F" = paste0("/", unit), "Clss_F"  = paste0("/", unit),
                 "Clss" = paste0("/", unit), "Dose" = paste0("/", unit))
  if (!is.na(normUnits[param])) {
    return(normUnits[[param]])
  } else {
    return("")
  }
}

################################################################################
# Remplacement in the doc
################################################################################

#-------------------------------------------------------------------------------
# Replace a text in the doc
.replaceInDoc <- function(doc, PLACEHOLDER, value){

  if(!is.null(value)){
    #print(paste0(PLACEHOLDER, " - ", value))
    doc <- doc %>% body_replace_all_text(PLACEHOLDER, as.character(value), only_at_cursor = F, warn=FALSE, fixed=F)
    doc <- doc %>% footers_replace_all_text(PLACEHOLDER,as.character(value), only_at_cursor = F, warn=FALSE, fixed=F)
    suppressWarnings(doc <- doc %>% headers_replace_all_text(PLACEHOLDER, as.character(value), fixed=F))
  }
  return(doc)
}

.addTableReport <- function(my_doc, PLACEHOLDER, value){

  suppressWarnings(my_doc <- my_doc %>% cursor_reach(keyword = PLACEHOLDER) %>%
                     body_add_flextable(value, pos="on"))
  return(my_doc)
}

# .addPlotReport <- function(x, keyword, value, width = 6, height = 4, res=300){
#
#   bCursor <- cursor_reach(x, keyword=keyword)
#   if(!is.logical(bCursor)){
#     body_add_gg(x, value=value, width = width, height = height, res = res, style="Centered")
#     cursor_reach(x, keyword=keyword)
#     body_remove(x)
#   }
# }

.addPlotReport <- function(x, keyword, value, width = 6.5, height = 4.5, res=300) {

  bCursor <- cursor_reach(x, keyword=keyword)
  if(!is.logical(bCursor)){

    if(is.ggplot(value[[1]])){
      for(index in length(value):1){
       body_add_gg(x, value=value[[index]], width = width, height = height, res = res, style="Centered")
        cursor_reach(x, keyword=keyword)
      }
    }else{
      body_add_gg(x, value=value, width = width, height = height, res = res, style="Centered")
      cursor_reach(x, keyword=keyword)
    }
    body_remove(x)
  }

}

.removePlaceHolder <- function(x, keyword){

  bCursor <- cursor_reach(x, keyword=keyword)
  if(!is.logical(bCursor)){
    body_remove(x)
  }

}

#' Replace the placeholders defined in the template document
#'
#' @param templatePath (\emph{string}) Path of the template document with .docx
#'   extension.
#' @param objects (\emph{list}) List containing either: \itemize{ \item{if a
#'   global variable name is the same as a placeholder defined in the document,
#'   strings containing the names of variables containing objects (e.g.,
#'   `"PLOT_MEANOBS"`)} \item{strings with the names of placeholders pointing to
#'   objects (e.g., `"PLOT_MEANOBS" = plotMeanObs`)} \item{strings pointing to
#'   lists or lists that consist of three elements: a string equal to a
#'   placeholder name, or an object, and two integers, height and width of the
#'   plot object replacing the placeholder, if the object is a plot (e.g.,
#'   `list("PLOT_MEANOBS", 5, 5)`, `"PLOT_MEANOBS" = list(plotMeanObs, 5, 5)`).}
#'   }
#' @param savePath (\emph{string}) Path for saving the document, ending in
#'   .docx. Will overwrite the file with the same path, if it exists.
#'
#' @examples
#' \dontrun{
#' library(lixoftConnectors)
#'
#' initializeLixoftConnectors("pkanalix")
#' demoPath <- getDemoPath()
#' loadProject(paste0(demoPath, "/3.bioequivalence/project_crossover_bioequivalence.pkx"))
#' setNCASettings(
#'   partialAucTime = list(T, c(0, 24)),
#'   computedNCAParameters = c("AUCINF_obs", "Cmax", "Tmax", "Tlast")
#' )
#' runNCAEstimation()
#'
#' TABLE_NCA_SUMMARY <- getNCASummaryTable(
#'   paramListSummary = c("AUCINF_obs", "Cmax", "AUC_0_24", "Tmax", "Tlast"),
#'   asMedian = c("Tmax", "Tlast"),
#'   stat = c("Mean", "CV%", "HMean"),
#'   nbSignDigit = 4,
#'   horizCatName = "FORM",
#'   renameHorizCat = c("ref" = "Reference", "test" = "Test"),
#'   vertCatName = "SEQ",
#'   filterID = c("1", "2")
#' )
#' PLOT_BOXPLOT_CMAX <- getBOXPLOT(
#'   param = "Cmax",
#'   splitCat = "FORM",
#'   renameSplitCat = c("ref" = "Reference", "test" = "Test"),
#'   addPoint = F
#' )
#' templatePath <- system.file("extdata", "Report_template_SAD.docx", package = "RsPKanalix")
#' replacePlaceholder(
#'   templatePath = templatePath,
#'   list("TABLE_NCA_SUMMARY",
#'        list("PLOT_BOXPLOT_CMAX", 5, 5),
#'        "DRUG_NAME" = "Theophylline 20 mg"),
#'   savePath = "generated_report.docx"
#' )
#' }
#'
#' @export
replacePlaceholder <- function(templatePath = NULL, objects, savePath = NULL) {
  if (!is.null(templatePath)) {
    doc <- .readTemplate(templatePath)
  } else {
    message("ERROR: templatePath needs to be assigned.")
    return(invisible(FALSE))
  }
  if (is.null(savePath)) {
    message("ERROR: savePath needs to be a string.")
    return(invisible(FALSE))
  }
  if (!dir.exists(dirname(savePath))) {
    message("Creating the save file directory.")
    dir.create(dirname(savePath))
  }

  nbObs <- length(objects)
  if (nbObs > 0) {
    for (indexObjects in 1:nbObs) {
      wd = 6.5; ht = 4;
      if (is.list(objects[[indexObjects]])) {
        if (length(objects[[indexObjects]]) != 3) {
          message(paste0("WARNING: object list must be of size 3 (name, width, height). Element '", objects[[indexObjects]][[1]], "' will be used as object with default width and height."))
          place <- objects[[indexObjects]][[1]]
        } else if (!is.numeric(objects[[indexObjects]][[2]]) || !is.numeric(objects[[indexObjects]][[3]])) {
          message(paste0("WARNING: width and height attributes must be numeric. Element '", objects[[indexObjects]][[1]], "' will be used as object with default width and height."))
          place <- objects[[indexObjects]][[1]]
        } else {
          place <- objects[[indexObjects]][[1]]
          wd <- objects[[indexObjects]][[2]]
          ht <- objects[[indexObjects]][[3]]
        }
      } else {
        place <- objects[[indexObjects]]
      }
      if (is.null(names(objects)) || names(objects)[indexObjects] == "") {
        object <- tryCatch(
          get(place),
          error = function(cond) {
            message(paste0("WARNING: Cannot find the object ", place, ", try providing the variable name using the = sign."))
            return(NA)
          }
        )
        if (any(is.na(object))) next
        placeholder <- place
      } else {
        object <- place
        placeholder <- names(objects)[indexObjects]
      }

      if (is.null(object)) {
        doc <- doc %>% .removePlaceHolder(placeholder)
      } else {
        if ((is.ggplot(object)) || (is.ggplot(object[[1]]))) {
          doc <- doc %>% .addPlotReport(placeholder, object, width = wd, height = ht)
        } else if (is.character(object)) {
          doc <- doc %>% .replaceInDoc(placeholder, object)
        } else {
          doc <- doc %>% .addTableReport(placeholder, object)
        }
      }
    }
  }
  .saveReport(doc, savePath)
  return(invisible(FALSE))
}

################################################################################
# Get the list of amounts in the description
################################################################################
.getAmountSTR <- function(cohortName, filterCohort = NULL) {

  cohorts <- .getDoseFromCohort(cohortName, filterCohort)[[1]]
  dosing <- .getDoseFromCohort(cohortName, filterCohort)[[2]]
  lenCohorts = length(cohorts)

  if (lenCohorts == 0) {
    doseSTR = dosing[1]
  } else if (lenCohorts == 1) {
    doseSTR = dosing[1]
  } else {
    doseSTR = NULL
    for (indexCohort in 1:lenCohorts) {
      doseSTR = paste0(doseSTR, dosing[indexCohort])
      if (indexCohort < lenCohorts - 1) {
        strLink = ", "
      } else if (indexCohort == (lenCohorts-1)) {
        if (lenCohorts == 2) {
          strLink = " and "
        } else {
          strLink = ", and "
        }
      } else {
        strLink = ""
      }
      doseSTR = paste0(doseSTR, strLink)
    }
  }

  return(paste0(doseSTR," ", .lixoftCall("getDataSettings")$units$dose))
}

################################################################################
# Get the global description
################################################################################
.getCohortDescription<-function(cohortName, filterCohort = NULL,  QD = F, addDose = T){

  if(QD){
    addTxt = " q.d."
  }else{
    addTxt = ""
  }
  doseU = paste0(.lixoftCall("getDataSettings")$units$dose, addTxt)

  cohorts <- .getDoseFromCohort(cohortName, filterCohort)[[1]]
  dosing <- .getDoseFromCohort(cohortName, filterCohort)[[2]]
  lenCohorts = length(cohorts)

  cohortSRT = ""
  if(lenCohorts==0){
    cohortSRT = paste0(cohortSRT, "Cohort ")
    if(addDose){
      cohortSRT = paste0(cohortSRT," (",dosing[1]," ",doseU,")")
    }
  }else if(lenCohorts==1){
    cohortSRT = paste0(cohortSRT, "Cohort ",cohorts[1])
    if(addDose){
      cohortSRT = paste0(cohortSRT," (",dosing[1]," ",doseU,")")
    }
  }else{
    cohortSRT = NULL
    for(indexCohort in 1:lenCohorts){
      cohortSRT = paste0(cohortSRT,"Cohort ",cohorts[indexCohort])
      if(addDose){
        cohortSRT = paste0(cohortSRT," (",dosing[indexCohort]," ",doseU,")")
      }
      if(indexCohort<lenCohorts-1){
        strLink = ", "
      }else if(indexCohort==(lenCohorts-1)){
        if(lenCohorts==2){
          strLink = " and "
        }else{
          strLink = ", and "
        }
      }else{
        strLink = ""
      }
      cohortSRT = paste0(cohortSRT, strLink)
    }
  }
  return(cohortSRT)
}

################################################################################
# Get the dose from a cohort
################################################################################
.getDoseFromCohort <-function(cohortName = NULL, filterCohort = NULL){

  doses <- NULL
  amounts <- .lixoftCall("getTreatmentsInformation")$amount
  if(is.null(cohortName)){
    doses <- unique(amounts)
    cohorts <- NULL
  }else{
    cat<- .updateModalities(cohortName, NULL, filterCohort)
    modCat <- cat[[1]];nbCat <- length(modCat)
    cov <- .lixoftCall("getCovariateInformation")$covariate[cohortName][[1]]
    cohorts <- modCat
    lenCohorts  = length(cohorts)
    doses <- NULL

    # Rank the cohorst by ascending doses
    for(indexCohort in 1:lenCohorts){
      doses <- c(doses, unique(amounts[which(cov==cohorts[indexCohort])]))
    }
    cohorts <- cohorts[sort.int(doses, index.return = T)$ix]
    doses <- doses[sort.int(doses, index.return = T)$ix]
  }

  return(list(cohorts, doses))
}

.saveReport <- function (document, path) {
  if (!endsWith(path, ".docx")) {
    path <- paste0(path, ".docx")
  }
  opened <- suppressWarnings(
    "try-error" %in% class(
      try(file(path,
               open = "w"),
          silent = TRUE
      )
    )
  )
  if (opened) {
    message("ERROR: save file is opened. Please close the file and try again.")
  } else {
    print(document, target = path)
    message(paste0("SUCCESS: file is successfully saved in '", path, "'."))
  }
}

.readTemplate <- function (path) {
  if (!file.exists(path)) {
    message("ERROR: template file doesn't exist. Check if the path is correct.")
  } else {
    return(read_docx(path))
  }
}
